/*  RAxML-VI-HPC (version 2.2) a program for sequential and parallel estimation of phylogenetic trees
 *  Copyright August 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *
 *  and
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  Alexandros.Stamatakis@epfl.ch
 *
 *  When publishing work that is based on the results from RAxML-VI-HPC please cite:
 *
 *  Alexandros Stamatakis:"RAxML-VI-HPC: maximum likelihood-based phylogenetic analyses with thousands of taxa and mixed models".
 *  Bioinformatics 2006; doi: 10.1093/bioinformatics/btl446
 */

#ifndef WIN32
#include <unistd.h>
#endif

#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdint.h>
#include <limits.h>
#include "axml.h"

#ifdef __SIM_SSE3

#include <stdint.h>
#include <xmmintrin.h>
#include <pmmintrin.h>

const union __attribute__ ((aligned (16)))
{
       uint64_t i[2];
       __m128d m;
} absMask = {0x7fffffffffffffffULL , 0x7fffffffffffffffULL };


#endif


#ifdef _USE_PTHREADS
#include <pthread.h>
extern volatile int NumberOfThreads;
#endif


/* C-OPT this function below will also be called very frequently, but should not take much time
   to compute, nonetheless it will be interesting cache-wise */

static void makeP(double z1, double z2, double *rptr, double *EI,  double *EIGN, int numberOfCategories, double *left, double *right, int data)
{
  int i, j, k;

  switch(data)
    {
    case BINARY_DATA:
      {
	double d1, d2;

	for(i = 0; i < numberOfCategories; i++)
	  {
	    d1 = exp(rptr[i] * EIGN[0] * z1);
	    d2 = exp(rptr[i] * EIGN[0] * z2);

	    /*printf("EIGN %f %f %f %f %f\n", EIGN[0], z1, z2, EI[0], EI[1]);*/

	    for(j = 0; j < 2; j++)
	      {
		left[i * 4 + j * 2] = 1.0;
		right[i * 4 + j * 2] = 1.0;

		left[i * 4 + j * 2 + 1]  = d1 * EI[j];
		right[i * 4 + j * 2 + 1] = d2 * EI[j];

		/*printf("RR %f %f %f %f\n", 	left[i * 4 + j * 2], 	right[i * 4 + j * 2],	left[i * 4 + j * 2 + 1],	right[i * 4 + j * 2 + 1]);*/
	      }
	  }
      }
      break;
    case DNA_DATA:
      {
	double d1[3], d2[3];

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 3; j++)
	      {
		d1[j] = exp(rptr[i] * EIGN[j] * z1);
		d2[j] = exp(rptr[i] * EIGN[j] * z2);
	      }

	    for(j = 0; j < 4; j++)
	      {
		left[i * 16 + j * 4] = 1.0;
		right[i * 16 + j * 4] = 1.0;

		for(k = 0; k < 3; k++)
		  {
		    left[i * 16 + j * 4 + k + 1]  = d1[k] * EI[3 * j + k];
		    right[i * 16 + j * 4 + k + 1] = d2[k] * EI[3 * j + k];
		  }
	      }
	  }
      }
      break;
    case SECONDARY_DATA:
      {
	double lz1[15], lz2[15], d1[15], d2[15];

	for(i = 0; i < 15; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 15; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 16; j++)
	      {
		left[256 * i  + 16 * j] = 1.0;
		right[256 * i + 16 * j] = 1.0;

		for(k = 1; k < 16; k++)
		  {
		    left[256 * i + 16 * j + k]  = d1[k-1] * EI[15 * j + (k-1)];
		    right[256 * i + 16 * j + k] = d2[k-1] * EI[15 * j + (k-1)];
		  }
	      }
	  }
      }
      break;
    case SECONDARY_DATA_6:
      {
	double lz1[5], lz2[5], d1[5], d2[5];

	for(i = 0; i < 5; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 5; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 6; j++)
	      {
		left[36 * i  + 6 * j] = 1.0;
		right[36 * i + 6 * j] = 1.0;

		for(k = 1; k < 6; k++)
		  {
		    left[36 * i + 6 * j + k]  = d1[k-1] * EI[5 * j + (k-1)];
		    right[36 * i + 6 * j + k] = d2[k-1] * EI[5 * j + (k-1)];
		  }
	      }
	  }
      }
      break;
    case SECONDARY_DATA_7:
      {
	double lz1[6], lz2[6], d1[6], d2[6];

	for(i = 0; i < 6; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 6; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 7; j++)
	      {
		left[49 * i  + 7 * j] = 1.0;
		right[49 * i + 7 * j] = 1.0;

		for(k = 1; k < 7; k++)
		  {
		    left[49 * i + 7 * j + k]  = d1[k-1] * EI[6 * j + (k-1)];
		    right[49 * i + 7 * j + k] = d2[k-1] * EI[6 * j + (k-1)];
		  }
	      }
	  }
      }
      break;
    case AA_DATA:
      {
	double lz1[19], lz2[19], d1[19], d2[19];

	for(i = 0; i < 19; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 19; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 20; j++)
	      {
		left[400 * i  + 20 * j] = 1.0;
		right[400 * i + 20 * j] = 1.0;

		for(k = 1; k < 20; k++)
		  {
		    left[400 * i + 20 * j + k]  = d1[k-1] * EI[19 * j + (k-1)];
		    right[400 * i + 20 * j + k] = d2[k-1] * EI[19 * j + (k-1)];
		  }
	      }
	  }
      }
      break;
    default:
      assert(0);
    }

}



static void makeP_FLOAT(double z1, double z2, double *rptr, double *EI,  double *EIGN, int numberOfCategories, float *left, float *right, int data)
{
  int i, j, k;

  switch(data)
    {
    case DNA_DATA:
       {
	double d1[3], d2[3];

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 3; j++)
	      {
		d1[j] = exp(rptr[i] * EIGN[j] * z1);
		d2[j] = exp(rptr[i] * EIGN[j] * z2);
	      }

	    for(j = 0; j < 4; j++)
	      {
		left[i * 16 + j * 4] = 1.0;
		right[i * 16 + j * 4] = 1.0;

		for(k = 0; k < 3; k++)
		  {
		    left[i * 16 + j * 4 + k + 1]  = ((float)(d1[k] * EI[3 * j + k]));
		    right[i * 16 + j * 4 + k + 1] = ((float)(d2[k] * EI[3 * j + k]));
		  }
	      }
	  }
       }
       break;
    case BINARY_DATA:
      {
	double d1, d2;

	for(i = 0; i < numberOfCategories; i++)
	  {
	    d1 = exp(rptr[i] * EIGN[0] * z1);
	    d2 = exp(rptr[i] * EIGN[0] * z2);

	    for(j = 0; j < 2; j++)
	      {
		left[i * 4 + j * 2] = 1.0;
		right[i * 4 + j * 2] = 1.0;

		left[i * 4 + j * 2 + 1]  = ((float)(d1 * EI[j]));
		right[i * 4 + j * 2 + 1] = ((float)(d2 * EI[j]));
	      }
	  }
      }
      break;
    case SECONDARY_DATA:
      {
	double lz1[15], lz2[15], d1[15], d2[15];

	for(i = 0; i < 15; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 15; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 16; j++)
	      {
		left[256 * i  + 16 * j] = 1.0;
		right[256 * i + 16 * j] = 1.0;

		for(k = 1; k < 16; k++)
		  {
		    left[256 * i + 16 * j + k]  = ((float)(d1[k-1] * EI[15 * j + (k-1)]));
		    right[256 * i + 16 * j + k] = ((float)(d2[k-1] * EI[15 * j + (k-1)]));
		  }
	      }
	  }
      }
      break;
    case SECONDARY_DATA_6:
      {
	double lz1[5], lz2[5], d1[5], d2[5];

	for(i = 0; i < 5; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 5; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 6; j++)
	      {
		left[36 * i  + 6 * j] = 1.0;
		right[36 * i + 6 * j] = 1.0;

		for(k = 1; k < 6; k++)
		  {
		    left[36 * i + 6 * j + k]  = ((float)(d1[k-1] * EI[5 * j + (k-1)]));
		    right[36 * i + 6 * j + k] = ((float)(d2[k-1] * EI[5 * j + (k-1)]));
		  }
	      }
	  }
      }
      break;
    case SECONDARY_DATA_7:
      {
	double lz1[6], lz2[6], d1[6], d2[6];

	for(i = 0; i < 6; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 6; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 7; j++)
	      {
		left[49 * i  + 7 * j] = 1.0;
		right[49 * i + 7 * j] = 1.0;

		for(k = 1; k < 7; k++)
		  {
		    left[49 * i + 7 * j + k]  = ((float)(d1[k-1] * EI[6 * j + (k-1)]));
		    right[49 * i + 7 * j + k] = ((float)(d2[k-1] * EI[6 * j + (k-1)]));
		  }
	      }
	  }
      }
      break;
    case AA_DATA:
      {
	double lz1[19], lz2[19], d1[19], d2[19];

	for(i = 0; i < 19; i++)
	  {
	    lz1[i] = EIGN[i] * z1;
	    lz2[i] = EIGN[i] * z2;
	  }

	for(i = 0; i < numberOfCategories; i++)
	  {
	    for(j = 0; j < 19; j++)
	      {
		d1[j] = exp (rptr[i] * lz1[j]);
		d2[j] = exp (rptr[i] * lz2[j]);
	      }

	    for(j = 0; j < 20; j++)
	      {
		left[400 * i  + 20 * j] = 1.0;
		right[400 * i + 20 * j] = 1.0;

		for(k = 1; k < 20; k++)
		  {
		    left[400 * i + 20 * j + k]  = ((float)(d1[k-1] * EI[19 * j + (k-1)]));
		    right[400 * i + 20 * j + k] = ((float)(d2[k-1] * EI[19 * j + (k-1)]));
		  }
	      }
	  }
      }
      break;
    default:
      assert(0);
    }  
}



static void newviewGTRCAT_BINARY( int tipCase,  double *EV,  int *cptr,
				  double *x1_start,  double *x2_start,  double *x3_start,  double *tipVector,
				  int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				  int n,  double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le,
    *ri,
    *x1, *x2, *x3;
  double
    ump_x1, ump_x2, x1px2[2];
  int i, j, k, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &(tipVector[2 * tipX2[i]]);
	    x3 = &x3_start[2 * i];
	    /*printf("%f %f %f %f\n", x1[0], x1[1], x2[0], x2[1]);*/

	    le =  &left[cptr[i] * 4];
	    ri =  &right[cptr[i] * 4];

	    for(j = 0; j < 2; j++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;
		for(k = 0; k < 2; k++)
		  {
		    ump_x1 += x1[k] * le[j * 2 + k];
		    ump_x2 += x2[k] * ri[j * 2 + k];
		  }
		x1px2[j] = ump_x1 * ump_x2;
	      }

	    for(j = 0; j < 2; j++)
	      x3[j] = 0.0;

	    for(j = 0; j < 2; j++)
	      for(k = 0; k < 2; k++)
		x3[k] += x1px2[j] * EV[j * 2 + k];	   
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &x2_start[2 * i];
	    x3 = &x3_start[2 * i];
	    /*printf("%f %f %f %f\n", x1[0], x1[1], x2[0], x2[1]);*/
	    le =  &left[cptr[i] * 4];
	    ri =  &right[cptr[i] * 4];

	    for(j = 0; j < 2; j++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;
		for(k = 0; k < 2; k++)
		  {
		    ump_x1 += x1[k] * le[j * 2 + k];
		    ump_x2 += x2[k] * ri[j * 2 + k];
		  }
		x1px2[j] = ump_x1 * ump_x2;
	      }

	    for(j = 0; j < 2; j++)
	      x3[j] = 0.0;

	    for(j = 0; j < 2; j++)
	      for(k = 0; k < 2; k++)
		x3[k] +=  x1px2[j] *  EV[2 * j + k];	   

	    scale = 1;
	    for(j = 0; j < 2 && scale; j++)
	      scale = (x3[j] < minlikelihood && x3[j] > minusminlikelihood);

	    if(scale)
	      {
		for(j = 0; j < 2; j++)
		  x3[j] *= twotothe256;

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	       
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
	{
	  x1 = &x1_start[2 * i];
	  x2 = &x2_start[2 * i];
	  x3 = &x3_start[2 * i];

	  le = &left[cptr[i] * 4];
	  ri = &right[cptr[i] * 4];

	  for(j = 0; j < 2; j++)
	    {
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;
	      for(k = 0; k < 2; k++)
		{
		  ump_x1 += x1[k] * le[j * 2 + k];
		  ump_x2 += x2[k] * ri[j * 2 + k];
		}
	      x1px2[j] = ump_x1 * ump_x2;
	    }

	  for(j = 0; j < 2; j++)
	    x3[j] = 0.0;

	  for(j = 0; j < 2; j++)
	    for(k = 0; k < 2; k++)
	      x3[k] +=  x1px2[j] *  EV[2 * j + k];	  

	  scale = 1;
	  for(j = 0; j < 2 && scale; j++)
	    scale = (x3[j] < minlikelihood && x3[j] > minusminlikelihood);

	  if(scale)
	    {
	      for(j = 0; j < 2; j++)
		x3[j] *= twotothe256;

	      if(useFastScaling)
		addScale += wgt[i];
	      else
		ex3[i]  += 1;	   
	    }
	}
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}

static void newviewGTRGAMMA_BINARY(int tipCase,
				   double *x1_start, double *x2_start, double *x3_start,
				   double *EV, double *tipVector,
				   int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				   const int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling
				   )
{
  double
    *x1, *x2, *x3;
  double
    ump_x1,
    ump_x2,
    x1px2[4];
  int i, j, k, l, scale, addScale = 0;


  /* C-OPT figure out if we are at an inner node who has two tips/leaves
     as descendants TIP_TIP, a tip and another inner node as descendant
     TIP_INNER, or two inner nodes as descendants INNER_INNER */

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[2 * tipX1[i]]);
	    x2 = &(tipVector[2 * tipX2[i]]);
	    x3 = &x3_start[i * 8];

	    for(j = 0; j < 8; j++)
	      x3[j] = 0.0;

	    for (j = 0; j < 4; j++)
	      {
		for (k = 0; k < 2; k++)
		  {
		    ump_x1 = 0.0;
		    ump_x2 = 0.0;

		    for (l=0; l < 2; l++)
		      {
			ump_x1 += x1[l] * left[ j*4 + k*2 + l];
			ump_x2 += x2[l] * right[j*4 + k*2 + l];
		      }

		    x1px2[k] = ump_x1 * ump_x2;
		  }

		for(k = 0; k < 2; k++)
		  for (l = 0; l < 2; l++)
		    x3[j * 2 + l] +=  x1px2[k] * EV[2 * k + l];

	      }	   
	  }
      }
      break;
    case TIP_INNER:
      {
	 for (i = 0; i < n; i++)
	   {
	     x1 = &(tipVector[2 * tipX1[i]]);
	     x2 = &x2_start[i * 8];
	     x3 = &x3_start[i * 8];

	     for(j = 0; j < 8; j++)
	       x3[j] = 0.0;

	     for (j = 0; j < 4; j++)
	       {
		 for (k = 0; k < 2; k++)
		   {
		     ump_x1 = 0.0;
		     ump_x2 = 0.0;

		     for (l=0; l < 2; l++)
		       {
			 ump_x1 += x1[l] * left[ j*4 + k*2 + l];
			 ump_x2 += x2[j*2 + l] * right[j*4 + k*2 + l];
		       }

		     x1px2[k] = ump_x1 * ump_x2;
		   }

		 for(k = 0; k < 2; k++)
		   for (l = 0; l < 2; l++)
		     x3[j * 2 + l] +=  x1px2[k] * EV[2 * k + l];

	       }	    

	     scale = 1;
	     for(l = 0; scale && (l < 8); l++)
	       scale = (ABS(x3[l]) <  minlikelihood);

	     if(scale)
	       {
		 for (l=0; l < 8; l++)
		   x3[l] *= twotothe256;
		 
		 if(useFastScaling)
		   addScale += wgt[i];
		 else
		   ex3[i]  += 1;	       
	       }

	   }
      }
      break;
    case INNER_INNER:

      /* C-OPT here we don't do any pre-computations
	 This should be the most compute intensive loop of the three
	 cases here. If we have one or two tips as descendants
	 we can take a couple of shortcuts */


     for (i = 0; i < n; i++)
       {
	 x1 = &x1_start[i * 8];
	 x2 = &x2_start[i * 8];
	 x3 = &x3_start[i * 8];

	 for(j = 0; j < 8; j++)
	   x3[j] = 0.0;

	 for (j = 0; j < 4; j++)
	   {
	     for (k = 0; k < 2; k++)
	       {
		 ump_x1 = 0.0;
		 ump_x2 = 0.0;

		 for (l=0; l < 2; l++)
		   {
		     ump_x1 += x1[j*2 + l] * left[ j*4 + k*2 + l];
		     ump_x2 += x2[j*2 + l] * right[j*4 + k*2 + l];
		   }

		 x1px2[k] = ump_x1 * ump_x2;
	       }

	     for(k = 0; k < 2; k++)
	       for (l = 0; l < 2; l++)
		 x3[j * 2 + l] +=  x1px2[k] * EV[2 * k + l];

	   }
	 
	 scale = 1;
	 for(l = 0; scale && (l < 8); l++)
	   scale = (ABS(x3[l]) <  minlikelihood);


	 if(scale)
	   {
	     for (l=0; l<8; l++)
	       x3[l] *= twotothe256;

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;	  
	   }
       }
     break;

    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}




static void newviewGTRCAT_FLOAT( int tipCase,  float *EV,  int *cptr,
				 float *x1_start,  float *x2_start,  float *x3_start,  float *tipVector,
				 int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				 int n,  float *left, float *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  float
    *le,
    *ri,
    *x1, *x2, *x3;
  float
    ump_x1, ump_x2, x1px2[4];
  int i, j, k, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &(tipVector[4 * tipX2[i]]);
	    x3 = &x3_start[4 * i];

	    le =  &left[cptr[i] * 16];
	    ri =  &right[cptr[i] * 16];

	    for(j = 0; j < 4; j++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;
		for(k = 0; k < 4; k++)
		  {
		    ump_x1 += x1[k] * le[j * 4 + k];
		    ump_x2 += x2[k] * ri[j * 4 + k];
		  }
		x1px2[j] = ump_x1 * ump_x2;
	      }

	    for(j = 0; j < 4; j++)
	      x3[j] = 0.0;

	    for(j = 0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		x3[k] += x1px2[j] * EV[j * 4 + k];	    
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[4 * i];
	    x3 = &x3_start[4 * i];

	    le =  &left[cptr[i] * 16];
	    ri =  &right[cptr[i] * 16];

	    for(j = 0; j < 4; j++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;
		for(k = 0; k < 4; k++)
		  {
		    ump_x1 += x1[k] * le[j * 4 + k];
		    ump_x2 += x2[k] * ri[j * 4 + k];
		  }
		x1px2[j] = ump_x1 * ump_x2;
	      }

	    for(j = 0; j < 4; j++)
	      x3[j] = 0.0;

	    for(j = 0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		x3[k] +=  x1px2[j] *  EV[4 * j + k];	    

	    scale = 1;
	    for(j = 0; j < 4 && scale; j++)
	      scale = (x3[j] < minlikelihood_FLOAT && x3[j] > minusminlikelihood_FLOAT);

	    if(scale)
	      {
		for(j = 0; j < 4; j++)
		  x3[j] *= twotothe256_FLOAT;
		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	      
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
	{
	  x1 = &x1_start[4 * i];
	  x2 = &x2_start[4 * i];
	  x3 = &x3_start[4 * i];

	  le = &left[cptr[i] * 16];
	  ri = &right[cptr[i] * 16];

	  for(j = 0; j < 4; j++)
	    {
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;
	      for(k = 0; k < 4; k++)
		{
		  ump_x1 += x1[k] * le[j * 4 + k];
		  ump_x2 += x2[k] * ri[j * 4 + k];
		}
	      x1px2[j] = ump_x1 * ump_x2;
	    }

	  for(j = 0; j < 4; j++)
	    x3[j] = 0.0;

	  for(j = 0; j < 4; j++)
	    for(k = 0; k < 4; k++)
	      x3[k] +=  x1px2[j] *  EV[4 * j + k];	  

	  scale = 1;
	  for(j = 0; j < 4 && scale; j++)
	    scale = (x3[j] < minlikelihood_FLOAT && x3[j] > minusminlikelihood_FLOAT);

	  if(scale)
	    {
	      for(j = 0; j < 4; j++)
		x3[j] *= twotothe256_FLOAT;

	      if(useFastScaling)
		addScale += wgt[i];
	      else
		ex3[i]  += 1;	     
	    }
	}
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}

#ifndef __SIM_SSE3

static void newviewGTRCAT( int tipCase,  double *EV,  int *cptr,
			   double *x1_start,  double *x2_start,  double *x3_start,  double *tipVector,
			   int *ex3, unsigned char *tipX1, unsigned char *tipX2,
			   int n,  double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le,
    *ri,
    *x1, *x2, *x3;
  double
    ump_x1, ump_x2, x1px2[4];
  int i, j, k, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &(tipVector[4 * tipX2[i]]);
	    x3 = &x3_start[4 * i];

	    le =  &left[cptr[i] * 16];
	    ri =  &right[cptr[i] * 16];

	    for(j = 0; j < 4; j++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;
		for(k = 0; k < 4; k++)
		  {
		    ump_x1 += x1[k] * le[j * 4 + k];
		    ump_x2 += x2[k] * ri[j * 4 + k];
		  }
		x1px2[j] = ump_x1 * ump_x2;
	      }

	    for(j = 0; j < 4; j++)
	      x3[j] = 0.0;

	    for(j = 0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		x3[k] += x1px2[j] * EV[j * 4 + k];	   
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    x1 = &(tipVector[4 * tipX1[i]]);
	    x2 = &x2_start[4 * i];
	    x3 = &x3_start[4 * i];

	    le =  &left[cptr[i] * 16];
	    ri =  &right[cptr[i] * 16];

	    for(j = 0; j < 4; j++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;
		for(k = 0; k < 4; k++)
		  {
		    ump_x1 += x1[k] * le[j * 4 + k];
		    ump_x2 += x2[k] * ri[j * 4 + k];
		  }
		x1px2[j] = ump_x1 * ump_x2;
	      }

	    for(j = 0; j < 4; j++)
	      x3[j] = 0.0;

	    for(j = 0; j < 4; j++)
	      for(k = 0; k < 4; k++)
		x3[k] +=  x1px2[j] *  EV[4 * j + k];	   

	    scale = 1;
	    for(j = 0; j < 4 && scale; j++)
	      scale = (x3[j] < minlikelihood && x3[j] > minusminlikelihood);	    	   
	    	    
	    if(scale)
	      {		    
		for(j = 0; j < 4; j++)
		  x3[j] *= twotothe256;
		
		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;		
	      }	     
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
	{
	  x1 = &x1_start[4 * i];
	  x2 = &x2_start[4 * i];
	  x3 = &x3_start[4 * i];

	  le = &left[cptr[i] * 16];
	  ri = &right[cptr[i] * 16];

	  for(j = 0; j < 4; j++)
	    {
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;
	      for(k = 0; k < 4; k++)
		{
		  ump_x1 += x1[k] * le[j * 4 + k];
		  ump_x2 += x2[k] * ri[j * 4 + k];
		}
	      x1px2[j] = ump_x1 * ump_x2;
	    }

	  for(j = 0; j < 4; j++)
	    x3[j] = 0.0;

	  for(j = 0; j < 4; j++)
	    for(k = 0; k < 4; k++)
	      x3[k] +=  x1px2[j] *  EV[4 * j + k];
	
	  scale = 1;
	  for(j = 0; j < 4 && scale; j++)
	    scale = (x3[j] < minlikelihood && x3[j] > minusminlikelihood);

	  if(scale)
	    {		    
	      for(j = 0; j < 4; j++)
		x3[j] *= twotothe256;
	      
	      if(useFastScaling)
		addScale += wgt[i];
	      else
		ex3[i]  += 1;		
	    }	  
	}
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}

#else

static void newviewGTRCAT( int tipCase,  double *EV,  int *cptr,
                           double *x1_start,  double *x2_start,  double *x3_start,  double *tipVector,
                           int *ex3, unsigned char *tipX1, unsigned char *tipX2,
                           int n,  double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le,
    *ri,
    *x1, 
    *x2, 
    *x3,
    ump_x1, ump_x2,    
    x1px2[4]  __attribute__ ((aligned (16))),
    EV_t[16] __attribute__ ((aligned (16)));
  
  int 
    i, j, k, scale, addScale = 0;

  for(i = 0; i < 4; i++)
    for (j=0; j < 4; j++)
      EV_t[4 * j + i] = EV[4 * i + j];

  switch(tipCase)
    {
    case TIP_TIP:
      {	
        for (i = 0; i < n; i++)
          {	   
	    __m128d x1_k0, x1_k2, x2_k0, x2_k2, x1px2_j01, x1px2_j23, x3_k01, x3_k23;

            x1 = &(tipVector[4 * tipX1[i]]);
            x2 = &(tipVector[4 * tipX2[i]]);
            x3 = &x3_start[4 * i];	   

            le =  &left[cptr[i] * 16];
            ri =  &right[cptr[i] * 16];
	   
            //
            // multiply/add left and right sides
            //

            x1_k0 = _mm_load_pd( &x1[0] );	 
            x1_k2 = _mm_load_pd( &x1[2] );
            x2_k0 = _mm_load_pd( &x2[0] );
            x2_k2 = _mm_load_pd( &x2[2] );	             
	    
            {
	      __m128d 
		le_j0_k0, le_j0_k2, ri_j0_k0, ri_j0_k2, 
		le_j1_k0, le_j1_k2, ri_j1_k0, ri_j1_k2,
		ump_x1_j01, ump_x2_j01;

              le_j0_k0 = _mm_load_pd( &le[0*4 + 0]);
              le_j0_k2 = _mm_load_pd( &le[0*4 + 2]);
              ri_j0_k0 = _mm_load_pd( &ri[0*4 + 0]);
              ri_j0_k2 = _mm_load_pd( &ri[0*4 + 2]);
	     
              le_j0_k0 = _mm_mul_pd( le_j0_k0, x1_k0 );
              le_j0_k2 = _mm_mul_pd( le_j0_k2, x1_k2 );
              ri_j0_k0 = _mm_mul_pd( ri_j0_k0, x2_k0 );
              ri_j0_k2 = _mm_mul_pd( ri_j0_k2, x2_k2 );

              le_j1_k0 = _mm_load_pd( &le[1*4 + 0]);
              le_j1_k2 = _mm_load_pd( &le[1*4 + 2]);
              ri_j1_k0 = _mm_load_pd( &ri[1*4 + 0]);
              ri_j1_k2 = _mm_load_pd( &ri[1*4 + 2]);

              le_j1_k0 = _mm_mul_pd( le_j1_k0, x1_k0 );
              le_j1_k2 = _mm_mul_pd( le_j1_k2, x1_k2 );
              ri_j1_k0 = _mm_mul_pd( ri_j1_k0, x2_k0 );
              ri_j1_k2 = _mm_mul_pd( ri_j1_k2, x2_k2 );

	      ump_x1_j01 = _mm_hadd_pd( _mm_hadd_pd( le_j0_k0, le_j0_k2), _mm_hadd_pd( le_j1_k0, le_j1_k2));
              ump_x2_j01 = _mm_hadd_pd( _mm_hadd_pd( ri_j0_k0, ri_j0_k2), _mm_hadd_pd( ri_j1_k0, ri_j1_k2));
              

              x1px2_j01 = _mm_mul_pd( ump_x1_j01, ump_x2_j01 );
            }
            
	    
            {
	      __m128d 
		le_j2_k0, le_j2_k2, ri_j2_k0, ri_j2_k2, 
		le_j3_k0, le_j3_k2, ri_j3_k0, ri_j3_k2,
		ump_x1_j23, ump_x2_j23;

              le_j2_k0 = _mm_load_pd( &le[2*4 + 0]);
              le_j2_k2 = _mm_load_pd( &le[2*4 + 2]);
              ri_j2_k0 = _mm_load_pd( &ri[2*4 + 0]);
              ri_j2_k2 = _mm_load_pd( &ri[2*4 + 2]);

              le_j2_k0 = _mm_mul_pd( le_j2_k0, x1_k0 );
              le_j2_k2 = _mm_mul_pd( le_j2_k2, x1_k2 );
              ri_j2_k0 = _mm_mul_pd( ri_j2_k0, x2_k0 );
              ri_j2_k2 = _mm_mul_pd( ri_j2_k2, x2_k2 );


              le_j3_k0 = _mm_load_pd( &le[3*4 + 0]);
              le_j3_k2 = _mm_load_pd( &le[3*4 + 2]);
              ri_j3_k0 = _mm_load_pd( &ri[3*4 + 0]);
              ri_j3_k2 = _mm_load_pd( &ri[3*4 + 2]);
              le_j3_k0 = _mm_mul_pd( le_j3_k0, x1_k0 );
              le_j3_k2 = _mm_mul_pd( le_j3_k2, x1_k2 );
              ri_j3_k0 = _mm_mul_pd( ri_j3_k0, x2_k0 );
              ri_j3_k2 = _mm_mul_pd( ri_j3_k2, x2_k2 );             

               ump_x1_j23 = _mm_hadd_pd( _mm_hadd_pd( le_j2_k0, le_j2_k2), _mm_hadd_pd( le_j3_k0, le_j3_k2));
               ump_x2_j23 = _mm_hadd_pd( _mm_hadd_pd( ri_j2_k0, ri_j2_k2), _mm_hadd_pd( ri_j3_k0, ri_j3_k2));
              x1px2_j23 = _mm_mul_pd( ump_x1_j23, ump_x2_j23 );
            }


            _mm_store_pd( &x1px2[0], x1px2_j01 );
            _mm_store_pd( &x1px2[2], x1px2_j23 );

            //
            // multiply by EV_t
            //

           
            {
	      __m128d EV_t_k0_j01, EV_t_k0_j23, EV_t_k1_j01, EV_t_k1_j23;

              EV_t_k0_j01 = _mm_load_pd( &EV_t[4 * 0 + 0]);
              EV_t_k0_j23 = _mm_load_pd( &EV_t[4 * 0 + 2]);

              EV_t_k0_j01 = _mm_mul_pd( EV_t_k0_j01, x1px2_j01 );
              EV_t_k0_j23 = _mm_mul_pd( EV_t_k0_j23, x1px2_j23 );
              EV_t_k0_j01 = _mm_hadd_pd( EV_t_k0_j01, EV_t_k0_j23);

              EV_t_k1_j01 = _mm_load_pd( &EV_t[4 * 1 + 0]);
              EV_t_k1_j23 = _mm_load_pd( &EV_t[4 * 1 + 2]);

              EV_t_k1_j01 = _mm_mul_pd( EV_t_k1_j01, x1px2_j01 );
              EV_t_k1_j23 = _mm_mul_pd( EV_t_k1_j23, x1px2_j23 );
              EV_t_k1_j01 = _mm_hadd_pd( EV_t_k1_j01, EV_t_k1_j23);
              x3_k01 = _mm_hadd_pd( EV_t_k0_j01, EV_t_k1_j01 );

            }

           
            {
	      __m128d EV_t_k2_j01, EV_t_k2_j23, EV_t_k3_j01, EV_t_k3_j23;

              EV_t_k2_j01 = _mm_load_pd( &EV_t[4 * 2 + 0]);
              EV_t_k2_j23 = _mm_load_pd( &EV_t[4 * 2 + 2]);

              EV_t_k2_j01 = _mm_mul_pd( EV_t_k2_j01, x1px2_j01 );
              EV_t_k2_j23 = _mm_mul_pd( EV_t_k2_j23, x1px2_j23 );
              EV_t_k2_j01 = _mm_hadd_pd( EV_t_k2_j01, EV_t_k2_j23);

              EV_t_k3_j01 = _mm_load_pd( &EV_t[4 * 3 + 0]);
              EV_t_k3_j23 = _mm_load_pd( &EV_t[4 * 3 + 2]);

              EV_t_k3_j01 = _mm_mul_pd( EV_t_k3_j01, x1px2_j01 );
              EV_t_k3_j23 = _mm_mul_pd( EV_t_k3_j23, x1px2_j23 );
              EV_t_k3_j01 = _mm_hadd_pd( EV_t_k3_j01, EV_t_k3_j23);
              x3_k23 = _mm_hadd_pd( EV_t_k2_j01, EV_t_k3_j01 );

            }

            _mm_store_pd( &x3[0], x3_k01);
            _mm_store_pd( &x3[2], x3_k23);	   
          }
      }
      break;
    case TIP_INNER:
      {
        for (i = 0; i < n; i++)
          {	  
	    __m128d x1_k0, x1_k2, x2_k0, x2_k2, x1px2_j01, x1px2_j23, x3_k01, x3_k23;

            x1 = &(tipVector[4 * tipX1[i]]);
            x2 = &x2_start[4 * i];
            x3 = &x3_start[4 * i];

            le =  &left[cptr[i] * 16];
            ri =  &right[cptr[i] * 16];

            //
            // multiply/add left and right sides
            //

            x1_k0 = _mm_load_pd( &x1[0] );
            x1_k2 = _mm_load_pd( &x1[2] );
            x2_k0 = _mm_load_pd( &x2[0] );
            x2_k2 = _mm_load_pd( &x2[2] );                      

            {
	      __m128d 
		le_j0_k0, le_j0_k2, ri_j0_k0, ri_j0_k2, 
		le_j1_k0, le_j1_k2, ri_j1_k0, ri_j1_k2,
		ump_x1_j01, ump_x2_j01;

              le_j0_k0 = _mm_load_pd( &le[0*4 + 0]);
              le_j0_k2 = _mm_load_pd( &le[0*4 + 2]);
              ri_j0_k0 = _mm_load_pd( &ri[0*4 + 0]);
              ri_j0_k2 = _mm_load_pd( &ri[0*4 + 2]);

              le_j0_k0 = _mm_mul_pd( le_j0_k0, x1_k0 );
              le_j0_k2 = _mm_mul_pd( le_j0_k2, x1_k2 );
              ri_j0_k0 = _mm_mul_pd( ri_j0_k0, x2_k0 );
              ri_j0_k2 = _mm_mul_pd( ri_j0_k2, x2_k2 );

              le_j1_k0 = _mm_load_pd( &le[1*4 + 0]);
              le_j1_k2 = _mm_load_pd( &le[1*4 + 2]);
              ri_j1_k0 = _mm_load_pd( &ri[1*4 + 0]);
              ri_j1_k2 = _mm_load_pd( &ri[1*4 + 2]);

              le_j1_k0 = _mm_mul_pd( le_j1_k0, x1_k0 );
              le_j1_k2 = _mm_mul_pd( le_j1_k2, x1_k2 );
              ri_j1_k0 = _mm_mul_pd( ri_j1_k0, x2_k0 );
              ri_j1_k2 = _mm_mul_pd( ri_j1_k2, x2_k2 );
            
              ump_x1_j01 = _mm_hadd_pd( _mm_hadd_pd( le_j0_k0, le_j0_k2), _mm_hadd_pd( le_j1_k0, le_j1_k2));
              ump_x2_j01 = _mm_hadd_pd( _mm_hadd_pd( ri_j0_k0, ri_j0_k2), _mm_hadd_pd( ri_j1_k0, ri_j1_k2));

              x1px2_j01 = _mm_mul_pd( ump_x1_j01, ump_x2_j01 );
            }
          
            {
	      __m128d 
		le_j2_k0, le_j2_k2, ri_j2_k0, ri_j2_k2, 
		le_j3_k0, le_j3_k2, ri_j3_k0, ri_j3_k2,
		ump_x1_j23, ump_x2_j23;

              le_j2_k0 = _mm_load_pd( &le[2*4 + 0]);
              le_j2_k2 = _mm_load_pd( &le[2*4 + 2]);
              ri_j2_k0 = _mm_load_pd( &ri[2*4 + 0]);
              ri_j2_k2 = _mm_load_pd( &ri[2*4 + 2]);

              le_j2_k0 = _mm_mul_pd( le_j2_k0, x1_k0 );
              le_j2_k2 = _mm_mul_pd( le_j2_k2, x1_k2 );
              ri_j2_k0 = _mm_mul_pd( ri_j2_k0, x2_k0 );
              ri_j2_k2 = _mm_mul_pd( ri_j2_k2, x2_k2 );

              le_j3_k0 = _mm_load_pd( &le[3*4 + 0]);
              le_j3_k2 = _mm_load_pd( &le[3*4 + 2]);
              ri_j3_k0 = _mm_load_pd( &ri[3*4 + 0]);
              ri_j3_k2 = _mm_load_pd( &ri[3*4 + 2]);

              le_j3_k0 = _mm_mul_pd( le_j3_k0, x1_k0 );
              le_j3_k2 = _mm_mul_pd( le_j3_k2, x1_k2 );
              ri_j3_k0 = _mm_mul_pd( ri_j3_k0, x2_k0 );
              ri_j3_k2 = _mm_mul_pd( ri_j3_k2, x2_k2 );

              ump_x1_j23 = _mm_hadd_pd( _mm_hadd_pd( le_j2_k0, le_j2_k2), _mm_hadd_pd( le_j3_k0, le_j3_k2));
              ump_x2_j23 = _mm_hadd_pd( _mm_hadd_pd( ri_j2_k0, ri_j2_k2), _mm_hadd_pd( ri_j3_k0, ri_j3_k2));
              x1px2_j23 = _mm_mul_pd( ump_x1_j23, ump_x2_j23 );
            }


            _mm_store_pd( &x1px2[0], x1px2_j01 );
            _mm_store_pd( &x1px2[2], x1px2_j23 );

            //
            // multiply by EV_t
            //


            
            {
	      __m128d EV_t_k0_j01, EV_t_k0_j23, EV_t_k1_j01, EV_t_k1_j23;

              EV_t_k0_j01 = _mm_load_pd( &EV_t[4 * 0 + 0]);
              EV_t_k0_j23 = _mm_load_pd( &EV_t[4 * 0 + 2]);

              EV_t_k0_j01 = _mm_mul_pd( EV_t_k0_j01, x1px2_j01 );
              EV_t_k0_j23 = _mm_mul_pd( EV_t_k0_j23, x1px2_j23 );
              EV_t_k0_j01 = _mm_hadd_pd( EV_t_k0_j01, EV_t_k0_j23);

              EV_t_k1_j01 = _mm_load_pd( &EV_t[4 * 1 + 0]);
              EV_t_k1_j23 = _mm_load_pd( &EV_t[4 * 1 + 2]);

              EV_t_k1_j01 = _mm_mul_pd( EV_t_k1_j01, x1px2_j01 );
              EV_t_k1_j23 = _mm_mul_pd( EV_t_k1_j23, x1px2_j23 );
              EV_t_k1_j01 = _mm_hadd_pd( EV_t_k1_j01, EV_t_k1_j23);
              x3_k01 = _mm_hadd_pd( EV_t_k0_j01, EV_t_k1_j01 );

            }
          
            {
	      __m128d EV_t_k2_j01, EV_t_k2_j23, EV_t_k3_j01, EV_t_k3_j23;

              EV_t_k2_j01 = _mm_load_pd( &EV_t[4 * 2 + 0]);
              EV_t_k2_j23 = _mm_load_pd( &EV_t[4 * 2 + 2]);

              EV_t_k2_j01 = _mm_mul_pd( EV_t_k2_j01, x1px2_j01 );
              EV_t_k2_j23 = _mm_mul_pd( EV_t_k2_j23, x1px2_j23 );
              EV_t_k2_j01 = _mm_hadd_pd( EV_t_k2_j01, EV_t_k2_j23);

              EV_t_k3_j01 = _mm_load_pd( &EV_t[4 * 3 + 0]);
              EV_t_k3_j23 = _mm_load_pd( &EV_t[4 * 3 + 2]);

              EV_t_k3_j01 = _mm_mul_pd( EV_t_k3_j01, x1px2_j01 );
              EV_t_k3_j23 = _mm_mul_pd( EV_t_k3_j23, x1px2_j23 );
              EV_t_k3_j01 = _mm_hadd_pd( EV_t_k3_j01, EV_t_k3_j23);
              x3_k23 = _mm_hadd_pd( EV_t_k2_j01, EV_t_k3_j01 );
            }

            _mm_store_pd( &x3[0], x3_k01);
            _mm_store_pd( &x3[2], x3_k23);
           

            scale = 1;
            for(j = 0; j < 4 && scale; j++)
              scale = (x3[j] < minlikelihood && x3[j] > minusminlikelihood);
	    
	    if(scale)
	      {
		for(j = 0; j < 4; j++)
		  x3[j] *= twotothe256;
		
		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i] += 1;
	      }	   
          }
      }
      break;
    case INNER_INNER:           
      for (i = 0; i < n; i++)
        {
	  __m128d x1_k0, x1_k2, x2_k0, x2_k2, x1px2_j01, x1px2_j23, x3_k01, x3_k23;

          x1 = &x1_start[4 * i];
          x2 = &x2_start[4 * i];
          x3 = &x3_start[4 * i];

          //
          // multiply/add left and right sides
          //

          x1_k0 = _mm_load_pd( &x1[0] );
          x1_k2 = _mm_load_pd( &x1[2] );
          x2_k0 = _mm_load_pd( &x2[0] );
          x2_k2 = _mm_load_pd( &x2[2] );

          le = &left[cptr[i] * 16];
          ri = &right[cptr[i] * 16];
          
          {
	    __m128d 
		le_j0_k0, le_j0_k2, ri_j0_k0, ri_j0_k2, 
		le_j1_k0, le_j1_k2, ri_j1_k0, ri_j1_k2,
		ump_x1_j01, ump_x2_j01;

            le_j0_k0 = _mm_load_pd( &le[0*4 + 0]);
	    le_j0_k2 = _mm_load_pd( &le[0*4 + 2]);
            ri_j0_k0 = _mm_load_pd( &ri[0*4 + 0]);
            ri_j0_k2 = _mm_load_pd( &ri[0*4 + 2]);

            le_j0_k0 = _mm_mul_pd( le_j0_k0, x1_k0 );
            le_j0_k2 = _mm_mul_pd( le_j0_k2, x1_k2 );
            ri_j0_k0 = _mm_mul_pd( ri_j0_k0, x2_k0 );
            ri_j0_k2 = _mm_mul_pd( ri_j0_k2, x2_k2 );

            le_j1_k0 = _mm_load_pd( &le[1*4 + 0]);
            le_j1_k2 = _mm_load_pd( &le[1*4 + 2]);
            ri_j1_k0 = _mm_load_pd( &ri[1*4 + 0]);
            ri_j1_k2 = _mm_load_pd( &ri[1*4 + 2]);

            le_j1_k0 = _mm_mul_pd( le_j1_k0, x1_k0 );
            le_j1_k2 = _mm_mul_pd( le_j1_k2, x1_k2 );
            ri_j1_k0 = _mm_mul_pd( ri_j1_k0, x2_k0 );
            ri_j1_k2 = _mm_mul_pd( ri_j1_k2, x2_k2 );
            
            ump_x1_j01 = _mm_hadd_pd( _mm_hadd_pd( le_j0_k0, le_j0_k2), _mm_hadd_pd( le_j1_k0, le_j1_k2));
            ump_x2_j01 = _mm_hadd_pd( _mm_hadd_pd( ri_j0_k0, ri_j0_k2), _mm_hadd_pd( ri_j1_k0, ri_j1_k2));

            x1px2_j01 = _mm_mul_pd( ump_x1_j01, ump_x2_j01 );
          }

        
          {
	    __m128d 
	      le_j2_k0, le_j2_k2, ri_j2_k0, ri_j2_k2, 
	      le_j3_k0, le_j3_k2, ri_j3_k0, ri_j3_k2,
	      ump_x1_j23, ump_x2_j23;
	    
	    le_j2_k0 = _mm_load_pd( &le[2*4 + 0]);
            le_j2_k2 = _mm_load_pd( &le[2*4 + 2]);
            ri_j2_k0 = _mm_load_pd( &ri[2*4 + 0]);
            ri_j2_k2 = _mm_load_pd( &ri[2*4 + 2]);

            le_j2_k0 = _mm_mul_pd( le_j2_k0, x1_k0 );
            le_j2_k2 = _mm_mul_pd( le_j2_k2, x1_k2 );
            ri_j2_k0 = _mm_mul_pd( ri_j2_k0, x2_k0 );
            ri_j2_k2 = _mm_mul_pd( ri_j2_k2, x2_k2 );

            le_j3_k0 = _mm_load_pd( &le[3*4 + 0]);
            le_j3_k2 = _mm_load_pd( &le[3*4 + 2]);
            ri_j3_k0 = _mm_load_pd( &ri[3*4 + 0]);
            ri_j3_k2 = _mm_load_pd( &ri[3*4 + 2]);

            le_j3_k0 = _mm_mul_pd( le_j3_k0, x1_k0 );
            le_j3_k2 = _mm_mul_pd( le_j3_k2, x1_k2 );
            ri_j3_k0 = _mm_mul_pd( ri_j3_k0, x2_k0 );
            ri_j3_k2 = _mm_mul_pd( ri_j3_k2, x2_k2 );           

            ump_x1_j23 = _mm_hadd_pd( _mm_hadd_pd( le_j2_k0, le_j2_k2), _mm_hadd_pd( le_j3_k0, le_j3_k2));
            ump_x2_j23 = _mm_hadd_pd( _mm_hadd_pd( ri_j2_k0, ri_j2_k2), _mm_hadd_pd( ri_j3_k0, ri_j3_k2));
            x1px2_j23 = _mm_mul_pd( ump_x1_j23, ump_x2_j23 );
          }


          _mm_store_pd( &x1px2[0], x1px2_j01 );
          _mm_store_pd( &x1px2[2], x1px2_j23 );

          //
          // multiply by EV_t
          //
        
          {
	    __m128d EV_t_k0_j01, EV_t_k0_j23, EV_t_k1_j01, EV_t_k1_j23;

            EV_t_k0_j01 = _mm_load_pd( &EV_t[4 * 0 + 0]);
            EV_t_k0_j23 = _mm_load_pd( &EV_t[4 * 0 + 2]);

            EV_t_k0_j01 = _mm_mul_pd( EV_t_k0_j01, x1px2_j01 );
            EV_t_k0_j23 = _mm_mul_pd( EV_t_k0_j23, x1px2_j23 );
            EV_t_k0_j01 = _mm_hadd_pd( EV_t_k0_j01, EV_t_k0_j23);

            EV_t_k1_j01 = _mm_load_pd( &EV_t[4 * 1 + 0]);
            EV_t_k1_j23 = _mm_load_pd( &EV_t[4 * 1 + 2]);

            EV_t_k1_j01 = _mm_mul_pd( EV_t_k1_j01, x1px2_j01 );
            EV_t_k1_j23 = _mm_mul_pd( EV_t_k1_j23, x1px2_j23 );
            EV_t_k1_j01 = _mm_hadd_pd( EV_t_k1_j01, EV_t_k1_j23);
            x3_k01 = _mm_hadd_pd( EV_t_k0_j01, EV_t_k1_j01 );
          }
         
          {
	    __m128d EV_t_k2_j01, EV_t_k2_j23, EV_t_k3_j01, EV_t_k3_j23;

            EV_t_k2_j01 = _mm_load_pd( &EV_t[4 * 2 + 0]);
            EV_t_k2_j23 = _mm_load_pd( &EV_t[4 * 2 + 2]);

            EV_t_k2_j01 = _mm_mul_pd( EV_t_k2_j01, x1px2_j01 );
            EV_t_k2_j23 = _mm_mul_pd( EV_t_k2_j23, x1px2_j23 );
            EV_t_k2_j01 = _mm_hadd_pd( EV_t_k2_j01, EV_t_k2_j23);

            EV_t_k3_j01 = _mm_load_pd( &EV_t[4 * 3 + 0]);
            EV_t_k3_j23 = _mm_load_pd( &EV_t[4 * 3 + 2]);

            EV_t_k3_j01 = _mm_mul_pd( EV_t_k3_j01, x1px2_j01 );
            EV_t_k3_j23 = _mm_mul_pd( EV_t_k3_j23, x1px2_j23 );
            EV_t_k3_j01 = _mm_hadd_pd( EV_t_k3_j01, EV_t_k3_j23);
            x3_k23 = _mm_hadd_pd( EV_t_k2_j01, EV_t_k3_j01 );
          }
	  	 

	  /*{	          	  	     
	    __m128d v1 = _mm_and_pd(x3_k01, absMask.m);
	    v1 = _mm_cmplt_pd(v1, _mm_set1_pd( minlikelihood ));
	    int s1 = _mm_movemask_pd( v1 );

	    if(s1 == 3)
	      {
		__m128d v2 = _mm_and_pd(x3_k23, absMask.m);  	   
		v2 = _mm_cmplt_pd(v2, _mm_set1_pd( minlikelihood ));
		int s2 = _mm_movemask_pd( v2 );	     
	    
		if(s2 == 3)
		  { 	       
		    x3_k01 = _mm_mul_pd( x3_k01, _mm_set1_pd( twotothe256 ));
		    x3_k23 = _mm_mul_pd( x3_k23, _mm_set1_pd( twotothe256 ));
		    
		    if(useFastScaling)
		    addScale += wgt[i];
		    else
		    ex3[i]  += 1;		  
		  }
	      }	     
	      } */        	    
	  
	  _mm_store_pd( &x3[0], x3_k01);
	  _mm_store_pd( &x3[2], x3_k23);
          

          
	  scale = 1;
	  for(j = 0; j < 4 && scale; j++)
            scale = (x3[j] < minlikelihood && x3[j] > minusminlikelihood);
	  
	  if(scale)
	    {
	      for(j = 0; j < 4; j++)
		x3[j] *= twotothe256;

	      if(useFastScaling)
		addScale += wgt[i];
	      else
		ex3[i] += 1;
	    }		 	 	    

	  /*
	    SSE for scaling with GTRCAT did not yield any speedups
	    
	    __m128d minlikelihood_sse = _mm_set1_pd( minlikelihood );
	    __m128d minusminlikelihood_sse = _mm_set1_pd( minusminlikelihood );
	    
	    __m128d crp_01 = _mm_cmplt_pd( x3_k01, minlikelihood_sse );
	    __m128d crp_23 = _mm_cmplt_pd( x3_k23, minlikelihood_sse );
	    
	    __m128d crn_01 = _mm_cmpgt_pd( x3_k01, minusminlikelihood_sse );
	    __m128d crn_23 = _mm_cmpgt_pd( x3_k23, minusminlikelihood_sse );
	    
	    int mm_crp01 = _mm_movemask_pd( crp_01 );
	    int mm_crp23 = _mm_movemask_pd( crp_23 );
	    
	    int mm_crn01 = _mm_movemask_pd( crn_01 );
	    int mm_crn23 = _mm_movemask_pd( crn_23 );
	    	    	    
	    if( mm_crp01 == 3 && mm_crp23 == 3 && mm_crn01 == 3 && mm_crn23 == 3 )
            {
	    __m128d twotothe256_sse = _mm_set1_pd( twotothe256 );
	    x3_k01 = _mm_mul_pd( x3_k01, twotothe256_sse );
	    x3_k23 = _mm_mul_pd( x3_k23, twotothe256_sse );

	    if(useFastScaling)
	    addScale += wgt[i];
	    else
	    ex3[i]  += 1;	  
            }
	    
	    _mm_store_pd( &x3[0], x3_k01);
	    _mm_store_pd( &x3[2], x3_k23);
	  */
        }
      break;
    default:
      assert(0);
    }
  
  if(useFastScaling)
    *scalerIncrement = addScale;
}

#endif





static void newviewGTRGAMMA(int tipCase,
			    double *x1_start, double *x2_start, double *x3_start,
			    double *EV, double *tipVector,
			    int *ex3, unsigned char *tipX1, unsigned char *tipX2,
			    const int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling
			    )
{
  int i, j, k, l, scale, addScale = 0;
  double
    *x1,
    *x2,
    *x3,
    buf,
    ump_x1,
    ump_x2;

#ifndef __SIM_SSE3
  double x1px2[4];
#else
  
  double 
    x1px2[4] __attribute__ ((aligned (16))),
    EV_t[16] __attribute__ ((aligned (16)));    

  for(k = 0; k < 4; k++)
    for (l=0; l < 4; l++)
      EV_t[4 * l + k] = EV[4 * k + l];
#endif


  switch(tipCase)
    {
    case TIP_TIP:
      {
	double *uX1, umpX1[256], *uX2, umpX2[256];

#ifndef _SIM_SSE3
	for(i = 1; i < 16; i++)
	  {
	    x1 = &(tipVector[i * 4]);

	    for(j=0; j<4; j++)
	      for(k=0; k<4; k++)
		{
		  umpX1[i*16 + j*4 + k] = 0.0;
		  umpX2[i*16 + j*4 + k] = 0.0;

		  for (l=0; l < 4; l++)
		    {
		      umpX1[i*16 + j*4 + k] += x1[l] * left[j*16 + k*4 + l];
		      umpX2[i*16 + j*4 + k] += x1[l] * right[j*16 + k*4 + l];
		    }
		}
	  }
#else
	for (i = 1; i < 16; i++)
	  {
	    __m128d x1_1 = _mm_load_pd(&(tipVector[i*4]));
	    __m128d x1_2 = _mm_load_pd(&(tipVector[i*4 + 2]));	   

	    for (j = 0; j < 4; j++)
	      for (k = 0; k < 4; k++)
		{		 
		  __m128d left1 = _mm_load_pd(&left[j*16 + k*4]);
		  __m128d left2 = _mm_load_pd(&left[j*16 + k*4 + 2]);
		  
		  __m128d acc = _mm_setzero_pd();

		  acc = _mm_add_pd(acc, _mm_mul_pd(left1, x1_1));
		  acc = _mm_add_pd(acc, _mm_mul_pd(left2, x1_2));
		  		  
		  acc = _mm_hadd_pd(acc, acc);
		  _mm_storel_pd(&umpX1[i*16 + j*4 + k], acc);
		}
	  
	    for (j = 0; j < 4; j++)
	      for (k = 0; k < 4; k++)
		{
		  __m128d left1 = _mm_load_pd(&right[j*16 + k*4]);
		  __m128d left2 = _mm_load_pd(&right[j*16 + k*4 + 2]);
		  
		  __m128d acc = _mm_setzero_pd();

		  acc = _mm_add_pd(acc, _mm_mul_pd(left1, x1_1));
		  acc = _mm_add_pd(acc, _mm_mul_pd(left2, x1_2));
		  		  
		  acc = _mm_hadd_pd(acc, acc);
		  _mm_storel_pd(&umpX2[i*16 + j*4 + k], acc);
		 
		}
	  }
#endif

	for (i = 0; i < n; i++)
	  {
	    x3 = &x3_start[i * 16];

	    uX1 = &umpX1[16 * tipX1[i]];
	    uX2 = &umpX2[16 * tipX2[i]];

	    for(j = 0; j < 16; j++)
	      x3[j] = 0.0;

	    for (j = 0; j < 4; j++)
	      for (k = 0; k < 4; k++)
		{
		  buf = uX1[j*4 + k] * uX2[j*4 + k];

		  for (l=0; l<4; l++)
		    x3[j * 4 + l] +=  buf * EV[4 * k + l];
		}	   
	  }
      }
      break;
    case TIP_INNER:
      {	
	double *uX1, umpX1[256];

#ifndef __SIM_SSE3
	for (i = 1; i < 16; i++)
	  {
	    x1 = &(tipVector[i*4]);

	    for (j = 0; j < 4; j++)
	      for (k = 0; k < 4; k++)
		{
		  umpX1[i*16 + j*4 + k] = 0.0;
		  for (l=0; l < 4; l++)
		    umpX1[i*16 + j*4 + k] += x1[l] * left[j*16 + k*4 + l];
		}
	  }
#else	
	for (i = 1; i < 16; i++)
	  {
	    __m128d x1_1 = _mm_load_pd(&(tipVector[i*4]));
	    __m128d x1_2 = _mm_load_pd(&(tipVector[i*4 + 2]));	   

	    for (j = 0; j < 4; j++)
	      for (k = 0; k < 4; k++)
		{		 
		  __m128d left1 = _mm_load_pd(&left[j*16 + k*4]);
		  __m128d left2 = _mm_load_pd(&left[j*16 + k*4 + 2]);
		  
		  __m128d acc = _mm_setzero_pd();

		  acc = _mm_add_pd(acc, _mm_mul_pd(left1, x1_1));
		  acc = _mm_add_pd(acc, _mm_mul_pd(left2, x1_2));
		  		  
		  acc = _mm_hadd_pd(acc, acc);
		  _mm_storel_pd(&umpX1[i*16 + j*4 + k], acc);		 
		}
	  }
#endif

	 for (i = 0; i < n; i++)
	   {
	     x2 = &x2_start[i * 16];
	     x3 = &x3_start[i * 16];

	     uX1 = &umpX1[16 * tipX1[i]];

	     for(j = 0; j < 16; j++)
	       x3[j] = 0.0;

	     for (j = 0; j < 4; j++)
	       {
#ifndef __SIM_SSE3

		 for (k = 0; k < 4; k++)
		   {
		     ump_x2 = 0.0;

		     for (l=0; l<4; l++)
		       ump_x2 += x2[j*4 + l] * right[j* 16 + k*4 + l];
		     x1px2[k] = uX1[j * 4 + k] * ump_x2;
		   }

		 for(k = 0; k < 4; k++)
		   for (l=0; l<4; l++)
		     x3[j * 4 + l] +=  x1px2[k] * EV[4 * k + l];

#else
		 //
		 // multiply/add right side
		 //
		 double *x2_p = &x2[j*4];
		 double *right_k0_p = &right[j*16];
		 double *right_k1_p = &right[j*16 + 1*4];
		 double *right_k2_p = &right[j*16 + 2*4];
		 double *right_k3_p = &right[j*16 + 3*4];
		 __m128d x2_0 = _mm_load_pd( &x2_p[0] );
		 __m128d x2_2 = _mm_load_pd( &x2_p[2] );

		 __m128d right_k0_0 = _mm_load_pd( &right_k0_p[0] );
		 __m128d right_k0_2 = _mm_load_pd( &right_k0_p[2] );
		 __m128d right_k1_0 = _mm_load_pd( &right_k1_p[0] );
		 __m128d right_k1_2 = _mm_load_pd( &right_k1_p[2] );
		 __m128d right_k2_0 = _mm_load_pd( &right_k2_p[0] );
		 __m128d right_k2_2 = _mm_load_pd( &right_k2_p[2] );
		 __m128d right_k3_0 = _mm_load_pd( &right_k3_p[0] );
		 __m128d right_k3_2 = _mm_load_pd( &right_k3_p[2] );



		 right_k0_0 = _mm_mul_pd( x2_0, right_k0_0);
		 right_k0_2 = _mm_mul_pd( x2_2, right_k0_2);

		 right_k1_0 = _mm_mul_pd( x2_0, right_k1_0);
		 right_k1_2 = _mm_mul_pd( x2_2, right_k1_2);

		 right_k0_0 = _mm_hadd_pd( right_k0_0, right_k0_2);
		 right_k1_0 = _mm_hadd_pd( right_k1_0, right_k1_2);
		 right_k0_0 = _mm_hadd_pd( right_k0_0, right_k1_0);


		 right_k2_0 = _mm_mul_pd( x2_0, right_k2_0);
		 right_k2_2 = _mm_mul_pd( x2_2, right_k2_2);


		 right_k3_0 = _mm_mul_pd( x2_0, right_k3_0);
		 right_k3_2 = _mm_mul_pd( x2_2, right_k3_2);

		 right_k2_0 = _mm_hadd_pd( right_k2_0, right_k2_2);
		 right_k3_0 = _mm_hadd_pd( right_k3_0, right_k3_2);
		 right_k2_0 = _mm_hadd_pd( right_k2_0, right_k3_0);

		 {
		   //
		   // load left side from tip vector
		   //
		   
		   __m128d uX1_k0_sse = _mm_load_pd( &uX1[j * 4] );
		   __m128d uX1_k2_sse = _mm_load_pd( &uX1[j * 4 + 2] );
		 
		 
		   //
		   // multiply left * right
		   //
		   
		   __m128d x1px2_k0 = _mm_mul_pd( uX1_k0_sse, right_k0_0 );
		   __m128d x1px2_k2 = _mm_mul_pd( uX1_k2_sse, right_k2_0 );
		   
		   
		   //
		   // multiply with EV matrix (!?)
		   //
		   
		   __m128d EV_t_l0_k0 = _mm_load_pd( &EV_t[4 * 0 + 0]);
		   __m128d EV_t_l0_k2 = _mm_load_pd( &EV_t[4 * 0 + 2]);
		   __m128d EV_t_l1_k0 = _mm_load_pd( &EV_t[4 * 1 + 0]);
		   __m128d EV_t_l1_k2 = _mm_load_pd( &EV_t[4 * 1 + 2]);
		   __m128d EV_t_l2_k0 = _mm_load_pd( &EV_t[4 * 2 + 0]);
		   __m128d EV_t_l2_k2 = _mm_load_pd( &EV_t[4 * 2 + 2]);
		   __m128d EV_t_l3_k0 = _mm_load_pd( &EV_t[4 * 3 + 0]);
		   __m128d EV_t_l3_k2 = _mm_load_pd( &EV_t[4 * 3 + 2]);
		   
		   EV_t_l0_k0 = _mm_mul_pd( x1px2_k0, EV_t_l0_k0 );
		   EV_t_l0_k2 = _mm_mul_pd( x1px2_k2, EV_t_l0_k2 );
		   EV_t_l0_k0 = _mm_hadd_pd( EV_t_l0_k0, EV_t_l0_k2 );
		   
		   EV_t_l1_k0 = _mm_mul_pd( x1px2_k0, EV_t_l1_k0 );
		   EV_t_l1_k2 = _mm_mul_pd( x1px2_k2, EV_t_l1_k2 );
		   
		   EV_t_l1_k0 = _mm_hadd_pd( EV_t_l1_k0, EV_t_l1_k2 );
		   EV_t_l0_k0 = _mm_hadd_pd( EV_t_l0_k0, EV_t_l1_k0 );
		   
		   EV_t_l2_k0 = _mm_mul_pd( x1px2_k0, EV_t_l2_k0 );
		   EV_t_l2_k2 = _mm_mul_pd( x1px2_k2, EV_t_l2_k2 );
		   EV_t_l2_k0 = _mm_hadd_pd( EV_t_l2_k0, EV_t_l2_k2 );
		   		   
		   EV_t_l3_k0 = _mm_mul_pd( x1px2_k0, EV_t_l3_k0 );
		   EV_t_l3_k2 = _mm_mul_pd( x1px2_k2, EV_t_l3_k2 );
		   EV_t_l3_k0 = _mm_hadd_pd( EV_t_l3_k0, EV_t_l3_k2 );
		   
		   EV_t_l2_k0 = _mm_hadd_pd( EV_t_l2_k0, EV_t_l3_k0 );
		   
		   _mm_store_pd( &x3[j * 4 + 0], EV_t_l0_k0 );
		   _mm_store_pd( &x3[j * 4 + 2], EV_t_l2_k0 );
		 }
#endif
	       }	     

	     scale = 1;
	     for(l = 0; scale && (l < 16); l++)
	       scale = (ABS(x3[l]) <  minlikelihood);

	     if(scale)
	       {
		 for (l=0; l<16; l++)
		   x3[l] *= twotothe256;

		 if(useFastScaling)
		   addScale += wgt[i];
		 else
		   ex3[i]  += 1;		 
	       }

	   }
      }
      break;
    case INNER_INNER:

      /* C-OPT here we don't do any pre-computations
	 This should be the most compute intensive loop of the three
	 cases here. If we have one or two tips as descendants
	 we can take a couple of shortcuts */


     for (i = 0; i < n; i++)
       {
	 x1 = &x1_start[i * 16];
	 x2 = &x2_start[i * 16];
	 x3 = &x3_start[i * 16];

	 for(j = 0; j < 16; j++)
	   x3[j] = 0.0;



	 for (j = 0; j < 4; j++)
	   {

#ifndef __SIM_SSE3
	     for (k = 0; k < 4; k++)
	       {
		 ump_x1 = 0.0;
		 ump_x2 = 0.0;

		 for (l=0; l<4; l++)
		   {
		     ump_x1 += x1[j*4 + l] * left[j*16 + k*4 +l];
		     ump_x2 += x2[j*4 + l] * right[j*16 + k*4 +l];
		   }




		 x1px2[k] = ump_x1 * ump_x2;
	       }

	     for(k = 0; k < 4; k++)
	       for (l=0; l<4; l++)
	         x3[j * 4 + l] +=  x1px2[k] * EV[4 * k + l];


#else	     
	     //
	     // multiply/add left side
	     //
	     
	     double *x1_p = &x1[j*4];
	     double *left_k0_p = &left[j*16];
	     double *left_k1_p = &left[j*16 + 1*4];
	     double *left_k2_p = &left[j*16 + 2*4];
	     double *left_k3_p = &left[j*16 + 3*4];
	     
	     __m128d x1_0 = _mm_load_pd( &x1_p[0] );
	     __m128d x1_2 = _mm_load_pd( &x1_p[2] );
	     
	     __m128d left_k0_0 = _mm_load_pd( &left_k0_p[0] );
	     __m128d left_k0_2 = _mm_load_pd( &left_k0_p[2] );
	     __m128d left_k1_0 = _mm_load_pd( &left_k1_p[0] );
	     __m128d left_k1_2 = _mm_load_pd( &left_k1_p[2] );
	     __m128d left_k2_0 = _mm_load_pd( &left_k2_p[0] );
	     __m128d left_k2_2 = _mm_load_pd( &left_k2_p[2] );
	     __m128d left_k3_0 = _mm_load_pd( &left_k3_p[0] );
	     __m128d left_k3_2 = _mm_load_pd( &left_k3_p[2] );
	     
	     left_k0_0 = _mm_mul_pd(x1_0, left_k0_0);
	     left_k0_2 = _mm_mul_pd(x1_2, left_k0_2);
	     
	     left_k1_0 = _mm_mul_pd(x1_0, left_k1_0);
	     left_k1_2 = _mm_mul_pd(x1_2, left_k1_2);
	     
	     left_k0_0 = _mm_hadd_pd( left_k0_0, left_k0_2 );
	     left_k1_0 = _mm_hadd_pd( left_k1_0, left_k1_2);
	     left_k0_0 = _mm_hadd_pd( left_k0_0, left_k1_0);
	     
	     left_k2_0 = _mm_mul_pd(x1_0, left_k2_0);
	     left_k2_2 = _mm_mul_pd(x1_2, left_k2_2);
	     
	     left_k3_0 = _mm_mul_pd(x1_0, left_k3_0);
	     left_k3_2 = _mm_mul_pd(x1_2, left_k3_2);
	     
	     left_k2_0 = _mm_hadd_pd( left_k2_0, left_k2_2);
	     left_k3_0 = _mm_hadd_pd( left_k3_0, left_k3_2);
	     left_k2_0 = _mm_hadd_pd( left_k2_0, left_k3_0);
	     
	     
	     //
	     // multiply/add right side
	     //
	     double *x2_p = &x2[j*4];
	     double *right_k0_p = &right[j*16];
	     double *right_k1_p = &right[j*16 + 1*4];
	     double *right_k2_p = &right[j*16 + 2*4];
	     double *right_k3_p = &right[j*16 + 3*4];
	     __m128d x2_0 = _mm_load_pd( &x2_p[0] );
	     __m128d x2_2 = _mm_load_pd( &x2_p[2] );
	     
	     __m128d right_k0_0 = _mm_load_pd( &right_k0_p[0] );
	     __m128d right_k0_2 = _mm_load_pd( &right_k0_p[2] );
	     __m128d right_k1_0 = _mm_load_pd( &right_k1_p[0] );
	     __m128d right_k1_2 = _mm_load_pd( &right_k1_p[2] );
	     __m128d right_k2_0 = _mm_load_pd( &right_k2_p[0] );
	     __m128d right_k2_2 = _mm_load_pd( &right_k2_p[2] );
	     __m128d right_k3_0 = _mm_load_pd( &right_k3_p[0] );
	     __m128d right_k3_2 = _mm_load_pd( &right_k3_p[2] );
	     
	     right_k0_0 = _mm_mul_pd( x2_0, right_k0_0);
	     right_k0_2 = _mm_mul_pd( x2_2, right_k0_2);
	     
	     right_k1_0 = _mm_mul_pd( x2_0, right_k1_0);
	     right_k1_2 = _mm_mul_pd( x2_2, right_k1_2);
	     
	     right_k0_0 = _mm_hadd_pd( right_k0_0, right_k0_2);
	     right_k1_0 = _mm_hadd_pd( right_k1_0, right_k1_2);
	     right_k0_0 = _mm_hadd_pd( right_k0_0, right_k1_0);
	     
	     right_k2_0 = _mm_mul_pd( x2_0, right_k2_0);
	     right_k2_2 = _mm_mul_pd( x2_2, right_k2_2);
	     
	     
	     right_k3_0 = _mm_mul_pd( x2_0, right_k3_0);
	     right_k3_2 = _mm_mul_pd( x2_2, right_k3_2);
	     
	     right_k2_0 = _mm_hadd_pd( right_k2_0, right_k2_2);
	     right_k3_0 = _mm_hadd_pd( right_k3_0, right_k3_2);
	     right_k2_0 = _mm_hadd_pd( right_k2_0, right_k3_0);	   

             //
             // multiply left * right
             //

	     __m128d x1px2_k0 = _mm_mul_pd( left_k0_0, right_k0_0 );
	     __m128d x1px2_k2 = _mm_mul_pd( left_k2_0, right_k2_0 );


             //
             // multiply with EV matrix (!?)
             //

	    __m128d EV_t_l0_k0 = _mm_load_pd( &EV_t[4 * 0 + 0]);
	    __m128d EV_t_l0_k2 = _mm_load_pd( &EV_t[4 * 0 + 2]);
	    __m128d EV_t_l1_k0 = _mm_load_pd( &EV_t[4 * 1 + 0]);
	    __m128d EV_t_l1_k2 = _mm_load_pd( &EV_t[4 * 1 + 2]);
	    __m128d EV_t_l2_k0 = _mm_load_pd( &EV_t[4 * 2 + 0]);
	    __m128d EV_t_l2_k2 = _mm_load_pd( &EV_t[4 * 2 + 2]);
	    __m128d EV_t_l3_k0 = _mm_load_pd( &EV_t[4 * 3 + 0]);
	    __m128d EV_t_l3_k2 = _mm_load_pd( &EV_t[4 * 3 + 2]);

	    EV_t_l0_k0 = _mm_mul_pd( x1px2_k0, EV_t_l0_k0 );
	    EV_t_l0_k2 = _mm_mul_pd( x1px2_k2, EV_t_l0_k2 );
	    EV_t_l0_k0 = _mm_hadd_pd( EV_t_l0_k0, EV_t_l0_k2 );

	    EV_t_l1_k0 = _mm_mul_pd( x1px2_k0, EV_t_l1_k0 );
	    EV_t_l1_k2 = _mm_mul_pd( x1px2_k2, EV_t_l1_k2 );

	    EV_t_l1_k0 = _mm_hadd_pd( EV_t_l1_k0, EV_t_l1_k2 );
	    EV_t_l0_k0 = _mm_hadd_pd( EV_t_l0_k0, EV_t_l1_k0 );

	    EV_t_l2_k0 = _mm_mul_pd( x1px2_k0, EV_t_l2_k0 );
	    EV_t_l2_k2 = _mm_mul_pd( x1px2_k2, EV_t_l2_k2 );
	    EV_t_l2_k0 = _mm_hadd_pd( EV_t_l2_k0, EV_t_l2_k2 );


	    EV_t_l3_k0 = _mm_mul_pd( x1px2_k0, EV_t_l3_k0 );
            EV_t_l3_k2 = _mm_mul_pd( x1px2_k2, EV_t_l3_k2 );
            EV_t_l3_k0 = _mm_hadd_pd( EV_t_l3_k0, EV_t_l3_k2 );

            EV_t_l2_k0 = _mm_hadd_pd( EV_t_l2_k0, EV_t_l3_k0 );

            _mm_store_pd( &x3[j * 4 + 0], EV_t_l0_k0 );
            _mm_store_pd( &x3[j * 4 + 2], EV_t_l2_k0 );
#endif
           }	 

	 scale = 1;
	 for(l = 0; scale && (l < 16); l++)
	   scale = (ABS(x3[l]) <  minlikelihood);

	 if(scale)
	   {
	     for (l=0; l<16; l++)
	       x3[l] *= twotothe256;

	      if(useFastScaling)
		addScale += wgt[i];
	      else
		ex3[i]  += 1;	   
	   }		      		
       }
   
     break;
    default:
      assert(0);
    }
  
  if(useFastScaling)
    *scalerIncrement = addScale;

}



static void newviewGTRGAMMA_FLOAT(int tipCase,
                                  float *x1_start, float *x2_start, float *x3_start,
                                  float *EV, float *tipVector,
                                  int *ex3, unsigned char *tipX1, unsigned char *tipX2,
                                  const int n, float *left, float *right, int *wgt, int *scalerIncrement, const boolean useFastScaling
                                  )
{
  float
    *x1, *x2, *x3;
  float
    buf,
    ump_x1,
    ump_x2;
  int i, j, k, l, scale, addScale = 0;

#ifndef __SIM_SSE3
  float x1px2[4];
#else
  float 
    x1px2[4] __attribute__ ((aligned (16))),
    EV_t[16] __attribute__ ((aligned (16)));

  for(k = 0; k < 4; k++)
    for (l=0; l < 4; l++)
      EV_t[4 * l + k] = EV[4 * k + l];
#endif

  switch(tipCase)
    {
    case TIP_TIP:
      {
#ifndef __SIM_SSE3
        float *uX1, umpX1[256], *uX2, umpX2[256];        
#else 
	float 
	  *uX1, 
	  umpX1[256] __attribute__ ((aligned (16))), 
	  *uX2, 
	  umpX2[256] __attribute__ ((aligned (16)));	
#endif

        for(i = 1; i < 16; i++)
          {
            x1 = &(tipVector[i * 4]);

            for(j=0; j<4; j++)
              for(k=0; k<4; k++)
                {
                  umpX1[i*16 + j*4 + k] = 0.0;
                  umpX2[i*16 + j*4 + k] = 0.0;

                  for (l=0; l < 4; l++)
                    {
                      umpX1[i*16 + j*4 + k] += x1[l] * left[j*16 + k*4 + l];
                      umpX2[i*16 + j*4 + k] += x1[l] * right[j*16 + k*4 + l];
                    }
                }
          }      

        for (i = 0; i < n; i++)
          {
            x3 = &x3_start[i * 16];

            uX1 = &umpX1[16 * tipX1[i]];
            uX2 = &umpX2[16 * tipX2[i]];

            for(j = 0; j < 16; j++)
              x3[j] = 0.0;

            for (j = 0; j < 4; j++)
              {
#ifndef __SIM_SSE3
                for (k = 0; k < 4; k++)
                  {
                    buf = uX1[j*4 + k] * uX2[j*4 + k];

                    for (l=0; l<4; l++)
                      x3[j * 4 + l] +=  buf * EV[4 * k + l];

                  }
#else
                //
                // load left side from tip vector
                //

                __m128 uX1_sse = _mm_load_ps( &uX1[j * 4] );


                //
                // load right side from tip vector
                //

                __m128 uX2_sse = _mm_load_ps( &uX2[j * 4] );

                //
                // multiply left * right
                //

                __m128 x1px2 = _mm_mul_ps( uX1_sse, uX2_sse );


                //
                // multiply with EV matrix (!?)
                //

                __m128 EV_t_l0 = _mm_load_ps( &EV_t[4 * 0]);
                __m128 EV_t_l1 = _mm_load_ps( &EV_t[4 * 1]);
                __m128 EV_t_l2 = _mm_load_ps( &EV_t[4 * 2]);
                __m128 EV_t_l3 = _mm_load_ps( &EV_t[4 * 3]);

                EV_t_l0 = _mm_mul_ps( x1px2, EV_t_l0 );
                EV_t_l1 = _mm_mul_ps( x1px2, EV_t_l1 );
                EV_t_l2 = _mm_mul_ps( x1px2, EV_t_l2 );
                EV_t_l3 = _mm_mul_ps( x1px2, EV_t_l3 );

                EV_t_l0 = _mm_hadd_ps( EV_t_l0, EV_t_l1 );
                EV_t_l2 = _mm_hadd_ps( EV_t_l2, EV_t_l3 );
                EV_t_l0 = _mm_hadd_ps( EV_t_l0, EV_t_l2 );

                _mm_store_ps( &x3[j * 4], EV_t_l0 );
#endif
              }            
          }
      }
      break;
    case TIP_INNER:
      {
#ifndef __SIM_SSE3	
        float *uX1, umpX1[256];       
#else
	float 
	  *uX1, 
	  umpX1[256] __attribute__ ((aligned (16)));       
#endif

        for (i = 1; i < 16; i++)
          {
            x1 = &(tipVector[i*4]);

            for (j = 0; j < 4; j++)
              for (k = 0; k < 4; k++)
                {
                  umpX1[i*16 + j*4 + k] = 0.0;
                  for (l=0; l < 4; l++)
                    umpX1[i*16 + j*4 + k] += x1[l] * left[j*16 + k*4 + l];
                }
          }       

         for (i = 0; i < n; i++)
           {
             x2 = &x2_start[i * 16];
             x3 = &x3_start[i * 16];

             uX1 = &umpX1[16 * tipX1[i]];

             for(j = 0; j < 16; j++)
               x3[j] = 0.0;

             for (j = 0; j < 4; j++)
               {

#ifndef __SIM_SSE3
                 for (k = 0; k < 4; k++)
                   {
                     ump_x2 = 0.0;

                     for (l=0; l<4; l++)
                       ump_x2 += x2[j*4 + l] * right[j* 16 + k*4 + l];
                     x1px2[k] = uX1[j * 4 + k] * ump_x2;
                   }

                 for(k = 0; k < 4; k++)
                   for (l=0; l<4; l++)
                     x3[j * 4 + l] +=  x1px2[k] * EV[4 * k + l];
#else
                 //
                 // multiply/add right side
                 //

                 float *right_k0_p = &right[j*16];
                 float *right_k1_p = &right[j*16 + 1*4];
                 float *right_k2_p = &right[j*16 + 2*4];
                 float *right_k3_p = &right[j*16 + 3*4];

                 float *x2_p = &x2[j*4];

                 __m128 x2 = _mm_load_ps( &x2_p[0] );
                 __m128 right_k0 = _mm_load_ps( &right_k0_p[0] );
                 __m128 right_k1 = _mm_load_ps( &right_k1_p[0] );
                 __m128 right_k2 = _mm_load_ps( &right_k2_p[0] );
                 __m128 right_k3 = _mm_load_ps( &right_k3_p[0] );


                 right_k0 = _mm_mul_ps( x2, right_k0);
                 right_k1 = _mm_mul_ps( x2, right_k1);

                 right_k2 = _mm_mul_ps( x2, right_k2);
                 right_k3 = _mm_mul_ps( x2, right_k3);

                 right_k0 = _mm_hadd_ps( right_k0, right_k1);
                 right_k2 = _mm_hadd_ps( right_k2, right_k3);
                 right_k0 = _mm_hadd_ps( right_k0, right_k2);


                 //
                 // load left side from tip vector
                 //

                 __m128 uX1_sse = _mm_load_ps( &uX1[j * 4] );


                 //
                 // multiply left * right
                 //

                 __m128 x1px2 = _mm_mul_ps( uX1_sse, right_k0 );


                 //
                 // multiply with EV matrix (!?)
                 //

                 __m128 EV_t_l0 = _mm_load_ps( &EV_t[4 * 0]);
                 __m128 EV_t_l1 = _mm_load_ps( &EV_t[4 * 1]);
                 __m128 EV_t_l2 = _mm_load_ps( &EV_t[4 * 2]);
                 __m128 EV_t_l3 = _mm_load_ps( &EV_t[4 * 3]);

                 EV_t_l0 = _mm_mul_ps( x1px2, EV_t_l0 );
                 EV_t_l1 = _mm_mul_ps( x1px2, EV_t_l1 );
                 EV_t_l2 = _mm_mul_ps( x1px2, EV_t_l2 );
                 EV_t_l3 = _mm_mul_ps( x1px2, EV_t_l3 );

                 EV_t_l0 = _mm_hadd_ps( EV_t_l0, EV_t_l1 );
                 EV_t_l2 = _mm_hadd_ps( EV_t_l2, EV_t_l3 );
                 EV_t_l0 = _mm_hadd_ps( EV_t_l0, EV_t_l2 );

                 _mm_store_ps( &x3[j * 4], EV_t_l0 );
#endif
               }             


#ifndef __SIM_SSE3

             scale = 1;
             for(l = 0; scale && (l < 16); l++)
               scale = (ABS(x3[l]) <  minlikelihood_FLOAT);

             if(scale)
               {
                 for (l=0; l<16; l++)
                   x3[l] *= twotothe256_FLOAT;

		 if(useFastScaling)
		   addScale += wgt[i];
		 else
		   ex3[i]  += 1;             
               }


#else
             __m128 zero = _mm_set1_ps(0.0);

             __m128 x3_0 = _mm_load_ps(&x3[4 * 0]);
             __m128 x3_1 = _mm_load_ps(&x3[4 * 1]);
             __m128 x3_2 = _mm_load_ps(&x3[4 * 2]);
             __m128 x3_3 = _mm_load_ps(&x3[4 * 3]);

             __m128 ax3_0 = _mm_sub_ps(zero, x3_0);
             __m128 ax3_1 = _mm_sub_ps(zero, x3_1);
             __m128 ax3_2 = _mm_sub_ps(zero, x3_2);
             __m128 ax3_3 = _mm_sub_ps(zero, x3_3);

             ax3_0 = _mm_max_ps(ax3_0, x3_0);
             ax3_1 = _mm_max_ps(ax3_1, x3_1);
             ax3_2 = _mm_max_ps(ax3_2, x3_2);
             ax3_3 = _mm_max_ps(ax3_3, x3_3);


             ax3_0 = _mm_max_ps( ax3_0, ax3_1 );
             ax3_2 = _mm_max_ps( ax3_2, ax3_3 );
             ax3_0 = _mm_max_ps( ax3_0, ax3_2 );

             __m128 minlikelihood_FLOAT_sse = _mm_set1_ps( minlikelihood_FLOAT );
             ax3_0 = _mm_cmplt_ps( ax3_0, minlikelihood_FLOAT_sse );

             int scale_bits = _mm_movemask_ps(ax3_0);


             if( scale_bits == 15 )
               {
                 __m128 twotothe256_FLOAT_sse = _mm_set1_ps( twotothe256_FLOAT );
                 x3_0 = _mm_mul_ps( x3_0, twotothe256_FLOAT_sse );
                 x3_1 = _mm_mul_ps( x3_1, twotothe256_FLOAT_sse );
                 x3_2 = _mm_mul_ps( x3_2, twotothe256_FLOAT_sse );
                 x3_3 = _mm_mul_ps( x3_3, twotothe256_FLOAT_sse );

                 _mm_store_ps( &x3[4 * 0], x3_0);
                 _mm_store_ps( &x3[4 * 1], x3_1);
                 _mm_store_ps( &x3[4 * 2], x3_2);
                 _mm_store_ps( &x3[4 * 3], x3_3);

		 if(useFastScaling)
		   addScale += wgt[i];
		 else
		   ex3[i]  += 1;              
               }

    #endif

           }
      }
      break;
    case INNER_INNER:    
     for (i = 0; i < n; i++)
       {
         x1 = &x1_start[i * 16];
         x2 = &x2_start[i * 16];
         x3 = &x3_start[i * 16];

         for(j = 0; j < 16; j++)
           x3[j] = 0.0;

         for (j = 0; j < 4; j++)
           {

#ifndef __SIM_SSE3
             for (k = 0; k < 4; k++)
               {
                 ump_x1 = 0.0;
                 ump_x2 = 0.0;

                 for (l=0; l<4; l++)
                   {
                     ump_x1 += x1[j*4 + l] * left[j*16 + k*4 +l];
                     ump_x2 += x2[j*4 + l] * right[j*16 + k*4 +l];
                   }

                 x1px2[k] = ump_x1 * ump_x2;
               }




             for(k = 0; k < 4; k++)
               for (l=0; l<4; l++)
                 x3[j * 4 + l] +=  x1px2[k] * EV[4 * k + l];


#else
             //
             // multiply/add left side
             //

             float *x1_p = &x1[j*4];

             float *left_k0_p = &left[j*16];
             float *left_k1_p = &left[j*16 + 1*4];
             float *left_k2_p = &left[j*16 + 2*4];
             float *left_k3_p = &left[j*16 + 3*4];


             __m128 x1 = _mm_load_ps( &x1_p[0] );
             __m128 left_k0 = _mm_load_ps( &left_k0_p[0] );
             __m128 left_k1 = _mm_load_ps( &left_k1_p[0] );
             __m128 left_k2 = _mm_load_ps( &left_k2_p[0] );
             __m128 left_k3 = _mm_load_ps( &left_k3_p[0] );

             left_k0 = _mm_mul_ps(x1, left_k0);
             left_k1 = _mm_mul_ps(x1, left_k1);

             left_k2 = _mm_mul_ps(x1, left_k2);
             left_k3 = _mm_mul_ps(x1, left_k3);

             left_k0 = _mm_hadd_ps( left_k0, left_k1);
             left_k2 = _mm_hadd_ps( left_k2, left_k3);
             left_k0 = _mm_hadd_ps( left_k0, left_k2);

             //
             // multiply/add right side
             //

             float *right_k0_p = &right[j*16];
             float *right_k1_p = &right[j*16 + 1*4];
             float *right_k2_p = &right[j*16 + 2*4];
             float *right_k3_p = &right[j*16 + 3*4];

             float *x2_p = &x2[j*4];

             __m128 x2 = _mm_load_ps( &x2_p[0] );
             __m128 right_k0 = _mm_load_ps( &right_k0_p[0] );
             __m128 right_k1 = _mm_load_ps( &right_k1_p[0] );
             __m128 right_k2 = _mm_load_ps( &right_k2_p[0] );
             __m128 right_k3 = _mm_load_ps( &right_k3_p[0] );


             right_k0 = _mm_mul_ps( x2, right_k0);
             right_k1 = _mm_mul_ps( x2, right_k1);

             right_k2 = _mm_mul_ps( x2, right_k2);
             right_k3 = _mm_mul_ps( x2, right_k3);

             right_k0 = _mm_hadd_ps( right_k0, right_k1);
             right_k2 = _mm_hadd_ps( right_k2, right_k3);
             right_k0 = _mm_hadd_ps( right_k0, right_k2);

             //
             // multiply left * right
             //

             __m128 x1px2 = _mm_mul_ps( left_k0, right_k0 );


             //
             // multiply with EV matrix (!?)
             //

             __m128 EV_t_l0 = _mm_load_ps( &EV_t[4 * 0]);
             __m128 EV_t_l1 = _mm_load_ps( &EV_t[4 * 1]);
             __m128 EV_t_l2 = _mm_load_ps( &EV_t[4 * 2]);
             __m128 EV_t_l3 = _mm_load_ps( &EV_t[4 * 3]);

             EV_t_l0 = _mm_mul_ps( x1px2, EV_t_l0 );
             EV_t_l1 = _mm_mul_ps( x1px2, EV_t_l1 );
             EV_t_l2 = _mm_mul_ps( x1px2, EV_t_l2 );
             EV_t_l3 = _mm_mul_ps( x1px2, EV_t_l3 );

             EV_t_l0 = _mm_hadd_ps( EV_t_l0, EV_t_l1 );
             EV_t_l2 = _mm_hadd_ps( EV_t_l2, EV_t_l3 );
             EV_t_l0 = _mm_hadd_ps( EV_t_l0, EV_t_l2 );


             _mm_store_ps( &x3[j * 4], EV_t_l0 );

#endif

           }

        
         scale = 1;


#ifndef __SIM_SSE3

         for(l = 0; scale && (l < 16); l++)
           scale = (ABS(x3[l]) <  minlikelihood_FLOAT);

         if(scale)
           {
             for (l=0; l<16; l++)
               x3[l] *= twotothe256_FLOAT;

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;            
           }
#else


         __m128 zero = _mm_set1_ps(0.0);

          __m128 x3_0 = _mm_load_ps(&x3[4 * 0]);
          __m128 x3_1 = _mm_load_ps(&x3[4 * 1]);
          __m128 x3_2 = _mm_load_ps(&x3[4 * 2]);
          __m128 x3_3 = _mm_load_ps(&x3[4 * 3]);

          __m128 ax3_0 = _mm_sub_ps(zero, x3_0);
          __m128 ax3_1 = _mm_sub_ps(zero, x3_1);
          __m128 ax3_2 = _mm_sub_ps(zero, x3_2);
          __m128 ax3_3 = _mm_sub_ps(zero, x3_3);

          ax3_0 = _mm_max_ps(ax3_0, x3_0);
          ax3_1 = _mm_max_ps(ax3_1, x3_1);
          ax3_2 = _mm_max_ps(ax3_2, x3_2);
          ax3_3 = _mm_max_ps(ax3_3, x3_3);


          ax3_0 = _mm_max_ps( ax3_0, ax3_1 );
          ax3_2 = _mm_max_ps( ax3_2, ax3_3 );
          ax3_0 = _mm_max_ps( ax3_0, ax3_2 );

          __m128 minlikelihood_FLOAT_sse = _mm_set1_ps( minlikelihood_FLOAT );
          ax3_0 = _mm_cmplt_ps( ax3_0, minlikelihood_FLOAT_sse );

//          unsigned int scaletest[4] __attribute__ ((aligned (16)));
//          _mm_store_ps( (float*) scaletest, ax3_0);

          int scale_bits = _mm_movemask_ps(ax3_0);

//          if(scaletest[0] && scaletest[1] && scaletest[2] && scaletest[3] )
          if( scale_bits == 15 )
            {
              __m128 twotothe256_FLOAT_sse = _mm_set1_ps( twotothe256_FLOAT );
              x3_0 = _mm_mul_ps( x3_0, twotothe256_FLOAT_sse );
              x3_1 = _mm_mul_ps( x3_1, twotothe256_FLOAT_sse );
              x3_2 = _mm_mul_ps( x3_2, twotothe256_FLOAT_sse );
              x3_3 = _mm_mul_ps( x3_3, twotothe256_FLOAT_sse );

              _mm_store_ps( &x3[4 * 0], x3_0);
              _mm_store_ps( &x3[4 * 1], x3_1);
              _mm_store_ps( &x3[4 * 2], x3_2);
              _mm_store_ps( &x3[4 * 3], x3_3);

	      if(useFastScaling)
		addScale += wgt[i];
	      else
		ex3[i]  += 1;            
            }

#endif
       }
     break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}




static void newviewGTRCATPROT(int tipCase, double *extEV,
			      int *cptr,
			      double *x1, double *x2, double *x3, double *tipVector,
			      int *ex3, unsigned char *tipX1, unsigned char *tipX2,
			      int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le, *ri, *v, *vl, *vr;
  double
    ump_x1, ump_x2, x1px2;
  int i, l, j, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 400];
	    ri = &right[cptr[i] * 400];

	    vl = &(tipVector[20 * tipX1[i]]);
	    vr = &(tipVector[20 * tipX2[i]]);
	    v  = &x3[20 * i];
#ifdef __SIM_SSE3
	    for(l = 0; l < 20; l+=2)
	      _mm_store_pd(&v[l], _mm_setzero_pd());	      		
#else
	    for(l = 0; l < 20; l++)
	      v[l] = 0.0;
#endif

	    for(l = 0; l < 20; l++)
	      {
#ifdef __SIM_SSE3
		__m128d x1v = _mm_setzero_pd();
		__m128d x2v = _mm_setzero_pd();	 
		double 
		  *ev = &extEV[l * 20],
		  *lv = &le[l * 20],
		  *rv = &ri[l * 20];

		for(j = 0; j < 20; j+=2)
		  {
		    x1v = _mm_add_pd(x1v, _mm_mul_pd(_mm_load_pd(&vl[j]), _mm_load_pd(&lv[j])));		    
		    x2v = _mm_add_pd(x2v, _mm_mul_pd(_mm_load_pd(&vr[j]), _mm_load_pd(&rv[j])));
		  }

		x1v = _mm_hadd_pd(x1v, x1v);
		x2v = _mm_hadd_pd(x2v, x2v);

		x1v = _mm_mul_pd(x1v, x2v);
		
		for(j = 0; j < 20; j+=2)
		  {
		    __m128d vv = _mm_load_pd(&v[j]);
		    vv = _mm_add_pd(vv, _mm_mul_pd(x1v, _mm_load_pd(&ev[j])));
		    _mm_store_pd(&v[j], vv);
		  }		    
#else
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 20; j++)
		  {
		    ump_x1 += vl[j] * le[l * 20 + j];
		    ump_x2 += vr[j] * ri[l * 20 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 20; j++)
		  v[j] += x1px2 * extEV[l * 20 + j];
#endif
	      }	   
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 400];
	    ri = &right[cptr[i] * 400];

	    vl = &(tipVector[20 * tipX1[i]]);
	    vr = &x2[20 * i];
	    v  = &x3[20 * i];

#ifdef __SIM_SSE3
	    for(l = 0; l < 20; l+=2)
	      _mm_store_pd(&v[l], _mm_setzero_pd());	      		
#else
	    for(l = 0; l < 20; l++)
	      v[l] = 0.0;
#endif
	   

	    for(l = 0; l < 20; l++)
	      {
#ifdef __SIM_SSE3

		__m128d x1v = _mm_setzero_pd();
		__m128d x2v = _mm_setzero_pd();	
		double 
		  *ev = &extEV[l * 20],
		  *lv = &le[l * 20],
		  *rv = &ri[l * 20];

		for(j = 0; j < 20; j+=2)
		  {
		    x1v = _mm_add_pd(x1v, _mm_mul_pd(_mm_load_pd(&vl[j]), _mm_load_pd(&lv[j])));		    
		    x2v = _mm_add_pd(x2v, _mm_mul_pd(_mm_load_pd(&vr[j]), _mm_load_pd(&rv[j])));
		  }

		x1v = _mm_hadd_pd(x1v, x1v);
		x2v = _mm_hadd_pd(x2v, x2v);

		x1v = _mm_mul_pd(x1v, x2v);
		
		for(j = 0; j < 20; j+=2)
		  {
		    __m128d vv = _mm_load_pd(&v[j]);
		    vv = _mm_add_pd(vv, _mm_mul_pd(x1v, _mm_load_pd(&ev[j])));
		    _mm_store_pd(&v[j], vv);
		  }		    
#else
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 20; j++)
		  {
		    ump_x1 += vl[j] * le[l * 20 + j];
		    ump_x2 += vr[j] * ri[l * 20 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 20; j++)
		  v[j] += x1px2 * extEV[l * 20 + j];
#endif
	      }
#ifdef __SIM_SSE3
	    { 	    
	      __m128d minlikelihood_sse = _mm_set1_pd( minlikelihood );
	      
	      scale = 1;
	      for(l = 0; scale && (l < 20); l += 2)
		{
		  __m128d vv = _mm_load_pd(&v[l]);
		  __m128d v1 = _mm_and_pd(vv, absMask.m);
		  v1 = _mm_cmplt_pd(v1,  minlikelihood_sse);
		  if(_mm_movemask_pd( v1 ) != 3)
		    scale = 0;
		}	    	  
	    }
#else
	    scale = 1;
	    for(l = 0; scale && (l < 20); l++)
	      scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));	   
#endif 

	    if(scale)
	      {
#ifdef __SIM_SSE3
		__m128d twoto = _mm_set_pd(twotothe256, twotothe256);

		for(l = 0; l < 20; l+=2)
		  {
		    __m128d ex3v = _mm_load_pd(&v[l]);
		    _mm_store_pd(&v[l], _mm_mul_pd(ex3v,twoto));		    
		  }
#else
		for(l = 0; l < 20; l++)
		  v[l] *= twotothe256;
#endif

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	      
	      }
	  }
      }
      break;
    case INNER_INNER:
      for(i = 0; i < n; i++)
	{
	  le = &left[cptr[i] * 400];
	  ri = &right[cptr[i] * 400];

	  vl = &x1[20 * i];
	  vr = &x2[20 * i];
	  v = &x3[20 * i];

#ifdef __SIM_SSE3
	    for(l = 0; l < 20; l+=2)
	      _mm_store_pd(&v[l], _mm_setzero_pd());	      		
#else
	    for(l = 0; l < 20; l++)
	      v[l] = 0.0;
#endif
	 
	  for(l = 0; l < 20; l++)
	    {
#ifdef __SIM_SSE3
		__m128d x1v = _mm_setzero_pd();
		__m128d x2v = _mm_setzero_pd();
		double 
		  *ev = &extEV[l * 20],
		  *lv = &le[l * 20],
		  *rv = &ri[l * 20];


		for(j = 0; j < 20; j+=2)
		  {
		    x1v = _mm_add_pd(x1v, _mm_mul_pd(_mm_load_pd(&vl[j]), _mm_load_pd(&lv[j])));		    
		    x2v = _mm_add_pd(x2v, _mm_mul_pd(_mm_load_pd(&vr[j]), _mm_load_pd(&rv[j])));
		  }

		x1v = _mm_hadd_pd(x1v, x1v);
		x2v = _mm_hadd_pd(x2v, x2v);

		x1v = _mm_mul_pd(x1v, x2v);
		
		for(j = 0; j < 20; j+=2)
		  {
		    __m128d vv = _mm_load_pd(&v[j]);
		    vv = _mm_add_pd(vv, _mm_mul_pd(x1v, _mm_load_pd(&ev[j])));
		    _mm_store_pd(&v[j], vv);
		  }		    
#else
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;

	      for(j = 0; j < 20; j++)
		{
		  ump_x1 += vl[j] * le[l * 20 + j];
		  ump_x2 += vr[j] * ri[l * 20 + j];
		}

	      x1px2 =  ump_x1 * ump_x2;

	      for(j = 0; j < 20; j++)
		v[j] += x1px2 * extEV[l * 20 + j];
#endif
	    }
#ifdef __SIM_SSE3
	    { 	    
	      __m128d minlikelihood_sse = _mm_set1_pd( minlikelihood );
	      
	      scale = 1;
	      for(l = 0; scale && (l < 20); l += 2)
		{
		  __m128d vv = _mm_load_pd(&v[l]);
		  __m128d v1 = _mm_and_pd(vv, absMask.m);
		  v1 = _mm_cmplt_pd(v1,  minlikelihood_sse);
		  if(_mm_movemask_pd( v1 ) != 3)
		    scale = 0;
		}	    	  
	    }
#else
	   scale = 1;
	   for(l = 0; scale && (l < 20); l++)
	     scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));
#endif	   

	   if(scale)
	     {
#ifdef __SIM_SSE3
	       __m128d twoto = _mm_set_pd(twotothe256, twotothe256);
	       
	       for(l = 0; l < 20; l+=2)
		 {
		   __m128d ex3v = _mm_load_pd(&v[l]);		  
		   _mm_store_pd(&v[l], _mm_mul_pd(ex3v,twoto));	
		 }		   		  
#else
	       for(l = 0; l < 20; l++)
		 v[l] *= twotothe256;
#endif

	       if(useFastScaling)
		 addScale += wgt[i];
	       else
		 ex3[i]  += 1;	      
	     }
	}
      break;
    default:
      assert(0);
    }
  
  if(useFastScaling)
    *scalerIncrement = addScale;

}


static void newviewGTRCATSECONDARY(int tipCase, double *extEV,
				   int *cptr,
				   double *x1, double *x2, double *x3, double *tipVector,
				   int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				   int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le, *ri, *v, *vl, *vr;
  double
    ump_x1, ump_x2, x1px2;
  int i, l, j, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 256];
	    ri = &right[cptr[i] * 256];

	    vl = &(tipVector[16 * tipX1[i]]);
	    vr = &(tipVector[16 * tipX2[i]]);
	    v  = &x3[16 * i];

	    for(l = 0; l < 16; l++)
	      v[l] = 0.0;

	    for(l = 0; l < 16; l++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 16; j++)
		  {
		    ump_x1 += vl[j] * le[l * 16 + j];
		    ump_x2 += vr[j] * ri[l * 16 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 16; j++)
		  v[j] += x1px2 * extEV[l * 16 + j];
	      }	    
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 256];
	    ri = &right[cptr[i] * 256];

	    vl = &(tipVector[16 * tipX1[i]]);
	    vr = &x2[16 * i];
	    v  = &x3[16 * i];

	    for(l = 0; l < 16; l++)
	      v[l] = 0.0;

	    for(l = 0; l < 16; l++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 16; j++)
		  {
		    ump_x1 += vl[j] * le[l * 16 + j];
		    ump_x2 += vr[j] * ri[l * 16 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 16; j++)
		  v[j] += x1px2 * extEV[l * 16 + j];
	      }

	    scale = 1;
	    for(l = 0; scale && (l < 16); l++)
	      scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));	    

	    if(scale)
	      {
		for(l = 0; l < 16; l++)
		  v[l] *= twotothe256;
		
		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	      
	      }
	  }
      }
      break;
    case INNER_INNER:
      for(i = 0; i < n; i++)
	{
	  le = &left[cptr[i] * 256];
	  ri = &right[cptr[i] * 256];

	  vl = &x1[16 * i];
	  vr = &x2[16 * i];
	  v = &x3[16 * i];

	  for(l = 0; l < 16; l++)
	    v[l] = 0.0;

	  for(l = 0; l < 16; l++)
	    {
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;

	      for(j = 0; j < 16; j++)
		{
		  ump_x1 += vl[j] * le[l * 16 + j];
		  ump_x2 += vr[j] * ri[l * 16 + j];
		}

	      x1px2 =  ump_x1 * ump_x2;

	      for(j = 0; j < 16; j++)
		v[j] += x1px2 * extEV[l * 16 + j];
	    }

	   scale = 1;
	   for(l = 0; scale && (l < 16); l++)
	     scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));
	  
	   if(scale)
	     {
	       for(l = 0; l < 16; l++)
		 v[l] *= twotothe256;

	       if(useFastScaling)
		 addScale += wgt[i];
	       else
		 ex3[i]  += 1;	     
	     }
	}
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}

static void newviewGTRCATSECONDARY_6(int tipCase, double *extEV,
				   int *cptr,
				   double *x1, double *x2, double *x3, double *tipVector,
				   int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				   int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le, *ri, *v, *vl, *vr;
  double
    ump_x1, ump_x2, x1px2;
  int i, l, j, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 36];
	    ri = &right[cptr[i] * 36];

	    vl = &(tipVector[6 * tipX1[i]]);
	    vr = &(tipVector[6 * tipX2[i]]);
	    v  = &x3[6 * i];

	    for(l = 0; l < 6; l++)
	      v[l] = 0.0;

	    for(l = 0; l < 6; l++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 6; j++)
		  {
		    ump_x1 += vl[j] * le[l * 6 + j];
		    ump_x2 += vr[j] * ri[l * 6 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 6; j++)
		  v[j] += x1px2 * extEV[l * 6 + j];
	      }	    
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 36];
	    ri = &right[cptr[i] * 36];

	    vl = &(tipVector[6 * tipX1[i]]);
	    vr = &x2[6 * i];
	    v  = &x3[6 * i];

	    for(l = 0; l < 6; l++)
	      v[l] = 0.0;

	    for(l = 0; l < 6; l++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 6; j++)
		  {
		    ump_x1 += vl[j] * le[l * 6 + j];
		    ump_x2 += vr[j] * ri[l * 6 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 6; j++)
		  v[j] += x1px2 * extEV[l * 6 + j];
	      }

	    scale = 1;
	    for(l = 0; scale && (l < 6); l++)
	      scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));	   

	    if(scale)
	      {
		for(l = 0; l < 6; l++)
		  v[l] *= twotothe256;

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	       
	      }
	  }
      }
      break;
    case INNER_INNER:
      for(i = 0; i < n; i++)
	{
	  le = &left[cptr[i] * 36];
	  ri = &right[cptr[i] * 36];

	  vl = &x1[6 * i];
	  vr = &x2[6 * i];
	  v = &x3[6 * i];

	  for(l = 0; l < 6; l++)
	    v[l] = 0.0;

	  for(l = 0; l < 6; l++)
	    {
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;

	      for(j = 0; j < 6; j++)
		{
		  ump_x1 += vl[j] * le[l * 6 + j];
		  ump_x2 += vr[j] * ri[l * 6 + j];
		}

	      x1px2 =  ump_x1 * ump_x2;

	      for(j = 0; j < 6; j++)
		v[j] += x1px2 * extEV[l * 6 + j];
	    }

	   scale = 1;
	   for(l = 0; scale && (l < 6); l++)
	     scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));	  

	   if(scale)
	     {
	       for(l = 0; l < 6; l++)
		 v[l] *= twotothe256;
	       
	       if(useFastScaling)
		 addScale += wgt[i];
	       else
		 ex3[i]  += 1;
	     }
	}
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}

static void newviewGTRCATSECONDARY_7(int tipCase, double *extEV,
				     int *cptr,
				     double *x1, double *x2, double *x3, double *tipVector,
				     int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				     int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double
    *le, *ri, *v, *vl, *vr;
  double
    ump_x1, ump_x2, x1px2;
  int i, l, j, scale, addScale = 0;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 49];
	    ri = &right[cptr[i] * 49];

	    vl = &(tipVector[7 * tipX1[i]]);
	    vr = &(tipVector[7 * tipX2[i]]);
	    v  = &x3[7 * i];

	    for(l = 0; l < 7; l++)
	      v[l] = 0.0;

	    for(l = 0; l < 7; l++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 7; j++)
		  {
		    ump_x1 += vl[j] * le[l * 7 + j];
		    ump_x2 += vr[j] * ri[l * 7 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 7; j++)
		  v[j] += x1px2 * extEV[l * 7 + j];
	      }	    
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    le = &left[cptr[i] * 49];
	    ri = &right[cptr[i] * 49];

	    vl = &(tipVector[7 * tipX1[i]]);
	    vr = &x2[7 * i];
	    v  = &x3[7 * i];

	    for(l = 0; l < 7; l++)
	      v[l] = 0.0;

	    for(l = 0; l < 7; l++)
	      {
		ump_x1 = 0.0;
		ump_x2 = 0.0;

		for(j = 0; j < 7; j++)
		  {
		    ump_x1 += vl[j] * le[l * 7 + j];
		    ump_x2 += vr[j] * ri[l * 7 + j];
		  }

		x1px2 = ump_x1 * ump_x2;

		for(j = 0; j < 7; j++)
		  v[j] += x1px2 * extEV[l * 7 + j];
	      }

	    scale = 1;
	    for(l = 0; scale && (l < 7); l++)
	      scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));	    

	    if(scale)
	      {
		for(l = 0; l < 7; l++)
		  v[l] *= twotothe256;

		if(useFastScaling)
		 addScale += wgt[i];
		else
		  ex3[i]  += 1;			     
	      }
	  }
      }
      break;
    case INNER_INNER:
      for(i = 0; i < n; i++)
	{
	  le = &left[cptr[i] * 49];
	  ri = &right[cptr[i] * 49];

	  vl = &x1[7 * i];
	  vr = &x2[7 * i];
	  v = &x3[7 * i];

	  for(l = 0; l < 7; l++)
	    v[l] = 0.0;

	  for(l = 0; l < 7; l++)
	    {
	      ump_x1 = 0.0;
	      ump_x2 = 0.0;

	      for(j = 0; j < 7; j++)
		{
		  ump_x1 += vl[j] * le[l * 7 + j];
		  ump_x2 += vr[j] * ri[l * 7 + j];
		}

	      x1px2 =  ump_x1 * ump_x2;

	      for(j = 0; j < 7; j++)
		v[j] += x1px2 * extEV[l * 7 + j];
	    }

	   scale = 1;
	   for(l = 0; scale && (l < 7); l++)
	     scale = ((v[l] < minlikelihood) && (v[l] > minusminlikelihood));	  

	   if(scale)
	     {
	       for(l = 0; l < 7; l++)
		 v[l] *= twotothe256;

	       if(useFastScaling)
		 addScale += wgt[i];
	       else
		 ex3[i]  += 1;	      
	     }
	}
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}



static void newviewGTRGAMMAPROT(int tipCase,
				double *x1, double *x2, double *x3, double *extEV, double *tipVector,
				int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double  *uX1, *uX2, *v;
  double x1px2;
  int  i, j, l, k, scale, addScale = 0;
  double *vl, *vr, al, ar;



  switch(tipCase)
    {
    case TIP_TIP:
      {
	double umpX1[1840], umpX2[1840];

	for(i = 0; i < 23; i++)
	  {
	    v = &(tipVector[20 * i]);

	    for(k = 0; k < 80; k++)
	      {
#ifdef __SIM_SSE3
		double *ll =  &left[k * 20];
		double *rr =  &right[k * 20];
		
		__m128d umpX1v = _mm_setzero_pd();
		__m128d umpX2v = _mm_setzero_pd();

		for(l = 0; l < 20; l+=2)
		  {
		    __m128d vv = _mm_load_pd(&v[l]);
		    umpX1v = _mm_add_pd(umpX1v, _mm_mul_pd(vv, _mm_load_pd(&ll[l])));
		    umpX2v = _mm_add_pd(umpX2v, _mm_mul_pd(vv, _mm_load_pd(&rr[l])));					
		  }
		
		umpX1v = _mm_hadd_pd(umpX1v, umpX1v);
		umpX2v = _mm_hadd_pd(umpX2v, umpX2v);
		
		_mm_storel_pd(&umpX1[80 * i + k], umpX1v);
		_mm_storel_pd(&umpX2[80 * i + k], umpX2v);
#else
		umpX1[80 * i + k] = 0.0;
		umpX2[80 * i + k] = 0.0;

		for(l = 0; l < 20; l++)
		  {
		    umpX1[80 * i + k] +=  v[l] *  left[k * 20 + l];
		    umpX2[80 * i + k] +=  v[l] * right[k * 20 + l];
		  }
#endif
	      }
	  }

	for(i = 0; i < n; i++)
	  {
	    uX1 = &umpX1[80 * tipX1[i]];
	    uX2 = &umpX2[80 * tipX2[i]];

	    for(j = 0; j < 4; j++)
	      {
		v = &x3[i * 80 + j * 20];

#ifdef __SIM_SSE3
		__m128d zero =  _mm_setzero_pd();
		for(k = 0; k < 20; k+=2)		  		    
		  _mm_store_pd(&v[k], zero);

		for(k = 0; k < 20; k++)
		  { 
		    double *eev = &extEV[k * 20];
		    x1px2 = uX1[j * 20 + k] * uX2[j * 20 + k];
		    __m128d x1px2v = _mm_set1_pd(x1px2);

		    for(l = 0; l < 20; l+=2)
		      {
		      	__m128d vv = _mm_load_pd(&v[l]);
			__m128d ee = _mm_load_pd(&eev[l]);

			vv = _mm_add_pd(vv, _mm_mul_pd(x1px2v,ee));
			
			_mm_store_pd(&v[l], vv);
		      }
		  }

#else

		for(k = 0; k < 20; k++)
		  v[k] = 0.0;

		for(k = 0; k < 20; k++)
		  {		   
		    x1px2 = uX1[j * 20 + k] * uX2[j * 20 + k];
		   
		    for(l = 0; l < 20; l++)		      					
		      v[l] += x1px2 * extEV[20 * k + l];		     
		  }
#endif
	      }	   
	  }
      }
      break;
    case TIP_INNER:
      {
	double umpX1[1840], ump_x2[20];


	for(i = 0; i < 23; i++)
	  {
	    v = &(tipVector[20 * i]);

	    for(k = 0; k < 80; k++)
	      {
#ifdef __SIM_SSE3
		double *ll =  &left[k * 20];
				
		__m128d umpX1v = _mm_setzero_pd();
		
		for(l = 0; l < 20; l+=2)
		  {
		    __m128d vv = _mm_load_pd(&v[l]);
		    umpX1v = _mm_add_pd(umpX1v, _mm_mul_pd(vv, _mm_load_pd(&ll[l])));		    					
		  }
		
		umpX1v = _mm_hadd_pd(umpX1v, umpX1v);				
		_mm_storel_pd(&umpX1[80 * i + k], umpX1v);		
#else	    
		umpX1[80 * i + k] = 0.0;

		for(l = 0; l < 20; l++)
		  umpX1[80 * i + k] +=  v[l] * left[k * 20 + l];
#endif

	      }
	  }

	for (i = 0; i < n; i++)
	  {
	    uX1 = &umpX1[80 * tipX1[i]];

	    for(k = 0; k < 4; k++)
	      {
		v = &(x2[80 * i + k * 20]);
#ifdef __SIM_SSE3	       
		for(l = 0; l < 20; l++)
		  {		   
		    double *r =  &right[k * 400 + l * 20];
		    __m128d ump_x2v = _mm_setzero_pd();	    
		    
		    for(j = 0; j < 20; j+= 2)
		      {
			__m128d vv = _mm_load_pd(&v[j]);
			__m128d rr = _mm_load_pd(&r[j]);
			ump_x2v = _mm_add_pd(ump_x2v, _mm_mul_pd(vv, rr));
		      }
		     
		    ump_x2v = _mm_hadd_pd(ump_x2v, ump_x2v);
		    
		    _mm_storel_pd(&ump_x2[l], ump_x2v);		   		     
		  }

		v = &(x3[80 * i + 20 * k]);

		__m128d zero =  _mm_setzero_pd();
		for(l = 0; l < 20; l+=2)		  		    
		  _mm_store_pd(&v[l], zero);
		  
		for(l = 0; l < 20; l++)
		  {
		    double *eev = &extEV[l * 20];
		    x1px2 = uX1[k * 20 + l]  * ump_x2[l];
		    __m128d x1px2v = _mm_set1_pd(x1px2);
		  
		    for(j = 0; j < 20; j+=2)
		      {
			__m128d vv = _mm_load_pd(&v[j]);
			__m128d ee = _mm_load_pd(&eev[j]);
			
			vv = _mm_add_pd(vv, _mm_mul_pd(x1px2v,ee));
			
			_mm_store_pd(&v[j], vv);
		      }		     		    
		  }			
#else
		for(l = 0; l < 20; l++)
		  {
		    ump_x2[l] = 0.0;

		    for(j = 0; j < 20; j++)
		      ump_x2[l] += v[j] * right[k * 400 + l * 20 + j];
		  }

		v = &(x3[80 * i + 20 * k]);

		for(l = 0; l < 20; l++)
		  v[l] = 0;

		for(l = 0; l < 20; l++)
		  {
		    x1px2 = uX1[k * 20 + l]  * ump_x2[l];
		    for(j = 0; j < 20; j++)
		      v[j] += x1px2 * extEV[l * 20  + j];
		  }
#endif
	      }
	   
#ifdef __SIM_SSE3
	    { 
	      v = &(x3[80 * i]);
	      __m128d minlikelihood_sse = _mm_set1_pd( minlikelihood );
	      
	      scale = 1;
	      for(l = 0; scale && (l < 80); l += 2)
		{
		  __m128d vv = _mm_load_pd(&v[l]);
		  __m128d v1 = _mm_and_pd(vv, absMask.m);
		  v1 = _mm_cmplt_pd(v1,  minlikelihood_sse);
		  if(_mm_movemask_pd( v1 ) != 3)
		    scale = 0;
		}	    	  
	    }
#else
	    v = &x3[80 * i];
	    scale = 1;
	    for(l = 0; scale && (l < 80); l++)
	      scale = (ABS(v[l]) <  minlikelihood);
#endif

	    if (scale)
	      {
#ifdef __SIM_SSE3
	       __m128d twoto = _mm_set_pd(twotothe256, twotothe256);
	       
	       for(l = 0; l < 80; l+=2)
		 {
		   __m128d ex3v = _mm_load_pd(&v[l]);		  
		   _mm_store_pd(&v[l], _mm_mul_pd(ex3v,twoto));	
		 }		   		  
#else
		for(l = 0; l < 80; l++)
		  v[l] *= twotothe256;
#endif

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	       
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
       {
	 for(k = 0; k < 4; k++)
	   {
	     vl = &(x1[80 * i + 20 * k]);
	     vr = &(x2[80 * i + 20 * k]);
	     v =  &(x3[80 * i + 20 * k]);

#ifdef __SIM_SSE3
	     __m128d zero =  _mm_setzero_pd();
	     for(l = 0; l < 20; l+=2)		  		    
	       _mm_store_pd(&v[l], zero);
#else
	     for(l = 0; l < 20; l++)
	       v[l] = 0;
#endif

	     for(l = 0; l < 20; l++)
	       {		 
#ifdef __SIM_SSE3
		 {
		   __m128d al = _mm_setzero_pd();
		   __m128d ar = _mm_setzero_pd();

		   double *ll   = &left[k * 400 + l * 20];
		   double *rr   = &right[k * 400 + l * 20];
		   double *EVEV = &extEV[20 * l];
		   
		   for(j = 0; j < 20; j+=2)
		     {
		       __m128d lv  = _mm_load_pd(&ll[j]);
		       __m128d rv  = _mm_load_pd(&rr[j]);
		       __m128d vll = _mm_load_pd(&vl[j]);
		       __m128d vrr = _mm_load_pd(&vr[j]);
		       
		       al = _mm_add_pd(al, _mm_mul_pd(vll, lv));
		       ar = _mm_add_pd(ar, _mm_mul_pd(vrr, rv));
		     }  		 
		       
		   al = _mm_hadd_pd(al, al);
		   ar = _mm_hadd_pd(ar, ar);
		   
		   al = _mm_mul_pd(al, ar);

		   for(j = 0; j < 20; j+=2)
		     {
		       __m128d vv  = _mm_load_pd(&v[j]);
		       __m128d EVV = _mm_load_pd(&EVEV[j]);

		       vv = _mm_add_pd(vv, _mm_mul_pd(al, EVV));

		       _mm_store_pd(&v[j], vv);
		     }		  		   		  
		 }		 
#else
		 al = 0.0;
		 ar = 0.0;

		 for(j = 0; j < 20; j++)
		   {
		     al += vl[j] * left[k * 400 + l * 20 + j];
		     ar += vr[j] * right[k * 400 + l * 20 + j];
		   }

		 x1px2 = al * ar;

		 for(j = 0; j < 20; j++)
		   v[j] += x1px2 * extEV[20 * l + j];
#endif
	       }
	   }
	 

#ifdef __SIM_SSE3
	 { 
	   v = &(x3[80 * i]);
	   __m128d minlikelihood_sse = _mm_set1_pd( minlikelihood );
	   
	   scale = 1;
	   for(l = 0; scale && (l < 80); l += 2)
	     {
	       __m128d vv = _mm_load_pd(&v[l]);
	       __m128d v1 = _mm_and_pd(vv, absMask.m);
	       v1 = _mm_cmplt_pd(v1,  minlikelihood_sse);
	       if(_mm_movemask_pd( v1 ) != 3)
		 scale = 0;
	     }	    	  
	 }
#else
	 v = &(x3[80 * i]);
	 scale = 1;
	 for(l = 0; scale && (l < 80); l++)
	   scale = ((ABS(v[l]) <  minlikelihood));
#endif

	 if (scale)
	   {
#ifdef __SIM_SSE3
	       __m128d twoto = _mm_set_pd(twotothe256, twotothe256);
	       
	       for(l = 0; l < 80; l+=2)
		 {
		   __m128d ex3v = _mm_load_pd(&v[l]);		  
		   _mm_store_pd(&v[l], _mm_mul_pd(ex3v,twoto));	
		 }		   		  
#else	     
	     for(l = 0; l < 80; l++)
	       v[l] *= twotothe256;
#endif

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;	  
	   }
       }
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}


static void newviewGTRGAMMAPROT_FLOAT(int tipCase,
				      float *x1, float *x2, float *x3, float *extEV, float *tipVector,
				      int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				      int n, float *left, float *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  float
    *uX1,
    *uX2,
    *v,
    x1px2,
    *vl,
    *vr,
    al,
    ar;

  int  i, j, l, k, scale, addScale = 0;



  switch(tipCase)
    {
    case TIP_TIP:
      {
	float umpX1[1840], umpX2[1840];

	for(i = 0; i < 23; i++)
	  {
	    v = &(tipVector[20 * i]);

	    for(k = 0; k < 80; k++)
	      {
		umpX1[80 * i + k] = 0.0;
		umpX2[80 * i + k] = 0.0;

		for(l = 0; l < 20; l++)
		  {
		    umpX1[80 * i + k] +=  v[l] *  left[k * 20 + l];
		    umpX2[80 * i + k] +=  v[l] * right[k * 20 + l];
		  }
	      }
	  }

	for(i = 0; i < n; i++)
	  {
	    uX1 = &umpX1[80 * tipX1[i]];
	    uX2 = &umpX2[80 * tipX2[i]];

	    for(j = 0; j < 4; j++)
	      {
		v = &x3[i * 80 + j * 20];

		for(k = 0; k < 20; k++)
		  v[k] = 0.0;

		for(k = 0; k < 20; k++)
		  {
		    x1px2 = uX1[j * 20 + k] * uX2[j * 20 + k];
		    for(l = 0; l < 20; l++)
		      v[l] += x1px2 * extEV[20 * k + l];
		  }
	      }	    
	  }
      }
      break;
    case TIP_INNER:
      {
	float umpX1[1840], ump_x2[20];


	for(i = 0; i < 23; i++)
	  {
	    v = &(tipVector[20 * i]);

	    for(k = 0; k < 80; k++)
	      {
		umpX1[80 * i + k] = 0.0;

		for(l = 0; l < 20; l++)
		  umpX1[80 * i + k] +=  v[l] * left[k * 20 + l];

	      }
	  }

	for (i = 0; i < n; i++)
	  {
	    uX1 = &umpX1[80 * tipX1[i]];

	    for(k = 0; k < 4; k++)
	      {
		v = &(x2[80 * i + k * 20]);
		for(l = 0; l < 20; l++)
		  {
		    ump_x2[l] = 0.0;

		    for(j = 0; j < 20; j++)
		      ump_x2[l] += v[j] * right[k * 400 + l * 20 + j];
		  }

		v = &(x3[80 * i + 20 * k]);

		for(l = 0; l < 20; l++)
		  v[l] = 0;

		for(l = 0; l < 20; l++)
		  {
		    x1px2 = uX1[k * 20 + l]  * ump_x2[l];
		    for(j = 0; j < 20; j++)
		      v[j] += x1px2 * extEV[l * 20  + j];
		  }
	      }
	    
	    v = &x3[80 * i];
	    scale = 1;
	    for(l = 0; scale && (l < 80); l++)
	      scale = (ABS(v[l]) <  minlikelihood_FLOAT);

	    if (scale)
	      {
		for(l = 0; l < 80; l++)
		  v[l] *= twotothe256_FLOAT;

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	      
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
       {
	 for(k = 0; k < 4; k++)
	   {
	     vl = &(x1[80 * i + 20 * k]);
	     vr = &(x2[80 * i + 20 * k]);
	     v =  &(x3[80 * i + 20 * k]);

	     for(l = 0; l < 20; l++)
	       v[l] = 0;

	     for(l = 0; l < 20; l++)
	       {
		 al = 0.0;
		 ar = 0.0;
		 for(j = 0; j < 20; j++)
		   {
		     al += vl[j] * left[k * 400 + l * 20 + j];
		     ar += vr[j] * right[k * 400 + l * 20 + j];
		   }

		 x1px2 = al * ar;
		 for(j = 0; j < 20; j++)
		   v[j] += x1px2 * extEV[20 * l + j];
	       }
	   }
	 
	 v = &(x3[80 * i]);
	 scale = 1;
	 for(l = 0; scale && (l < 80); l++)
	   scale = ((ABS(v[l]) <  minlikelihood_FLOAT));

	 if (scale)
	   {
	     for(l = 0; l < 80; l++)
	       v[l] *= twotothe256_FLOAT;

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;	   
	   }
       }
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}


static void newviewGTRGAMMASECONDARY(int tipCase,
				     double *x1, double *x2, double *x3, double *extEV, double *tipVector,
				     int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				     int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double  *v;
  double x1px2;
  int  i, j, l, k, scale, addScale = 0;
  double *vl, *vr, al, ar;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for(i = 0; i < n; i++)
	  {
	    for(k = 0; k < 4; k++)
	      {
		vl = &(tipVector[16 * tipX1[i]]);
		vr = &(tipVector[16 * tipX2[i]]);
		v =  &(x3[64 * i + 16 * k]);

		for(l = 0; l < 16; l++)
		  v[l] = 0;

		for(l = 0; l < 16; l++)
		  {
		    al = 0.0;
		    ar = 0.0;
		    for(j = 0; j < 16; j++)
		      {
			al += vl[j] * left[k * 256 + l * 16 + j];
			ar += vr[j] * right[k * 256 + l * 16 + j];
		      }

		    x1px2 = al * ar;
		    for(j = 0; j < 16; j++)
		      v[j] += x1px2 * extEV[16 * l + j];
		  }
	      }	    
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    for(k = 0; k < 4; k++)
	      {
		vl = &(tipVector[16 * tipX1[i]]);
		vr = &(x2[64 * i + 16 * k]);
		v =  &(x3[64 * i + 16 * k]);

		for(l = 0; l < 16; l++)
		  v[l] = 0;

		for(l = 0; l < 16; l++)
		  {
		    al = 0.0;
		    ar = 0.0;
		    for(j = 0; j < 16; j++)
		      {
			al += vl[j] * left[k * 256 + l * 16 + j];
			ar += vr[j] * right[k * 256 + l * 16 + j];
		      }

		    x1px2 = al * ar;
		    for(j = 0; j < 16; j++)
		      v[j] += x1px2 * extEV[16 * l + j];
		  }
	      }
	   
	    v = &x3[64 * i];
	    scale = 1;
	    for(l = 0; scale && (l < 64); l++)
	      scale = (ABS(v[l]) <  minlikelihood);

	    if (scale)
	      {
		for(l = 0; l < 64; l++)
		  v[l] *= twotothe256;

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	      
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
       {
	 for(k = 0; k < 4; k++)
	   {
	     vl = &(x1[64 * i + 16 * k]);
	     vr = &(x2[64 * i + 16 * k]);
	     v =  &(x3[64 * i + 16 * k]);

	     for(l = 0; l < 16; l++)
	       v[l] = 0;

	     for(l = 0; l < 16; l++)
	       {
		 al = 0.0;
		 ar = 0.0;
		 for(j = 0; j < 16; j++)
		   {
		     al += vl[j] * left[k * 256 + l * 16 + j];
		     ar += vr[j] * right[k * 256 + l * 16 + j];
		   }

		 x1px2 = al * ar;
		 for(j = 0; j < 16; j++)
		   v[j] += x1px2 * extEV[16 * l + j];
	       }
	   }
	 
	 v = &(x3[64 * i]);
	 scale = 1;
	 for(l = 0; scale && (l < 64); l++)
	   scale = ((ABS(v[l]) <  minlikelihood));

	 if (scale)
	   {
	     for(l = 0; l < 64; l++)
	       v[l] *= twotothe256;

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;	    
	   }
       }
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}


static void newviewGTRGAMMASECONDARY_6(int tipCase,
				       double *x1, double *x2, double *x3, double *extEV, double *tipVector,
				       int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				       int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double  *v;
  double x1px2;
  int  i, j, l, k, scale, addScale = 0;
  double *vl, *vr, al, ar;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for(i = 0; i < n; i++)
	  {
	    for(k = 0; k < 4; k++)
	      {
		vl = &(tipVector[6 * tipX1[i]]);
		vr = &(tipVector[6 * tipX2[i]]);
		v =  &(x3[24 * i + 6 * k]);

		for(l = 0; l < 6; l++)
		  v[l] = 0;

		for(l = 0; l < 6; l++)
		  {
		    al = 0.0;
		    ar = 0.0;
		    for(j = 0; j < 6; j++)
		      {
			al += vl[j] * left[k * 36 + l * 6 + j];
			ar += vr[j] * right[k * 36 + l * 6 + j];
		      }

		    x1px2 = al * ar;
		    for(j = 0; j < 6; j++)
		      v[j] += x1px2 * extEV[6 * l + j];
		  }
	      }	   
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    for(k = 0; k < 4; k++)
	      {
		vl = &(tipVector[6 * tipX1[i]]);
		vr = &(x2[24 * i + 6 * k]);
		v =  &(x3[24 * i + 6 * k]);

		for(l = 0; l < 6; l++)
		  v[l] = 0;

		for(l = 0; l < 6; l++)
		  {
		    al = 0.0;
		    ar = 0.0;
		    for(j = 0; j < 6; j++)
		      {
			al += vl[j] * left[k * 36 + l * 6 + j];
			ar += vr[j] * right[k * 36 + l * 6 + j];
		      }

		    x1px2 = al * ar;
		    for(j = 0; j < 6; j++)
		      v[j] += x1px2 * extEV[6 * l + j];
		  }
	      }
	   
	    v = &x3[24 * i];
	    scale = 1;
	    for(l = 0; scale && (l < 24); l++)
	      scale = (ABS(v[l]) <  minlikelihood);

	    if(scale)
	      {
		for(l = 0; l < 24; l++)
		  v[l] *= twotothe256;

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;		
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
       {
	 for(k = 0; k < 4; k++)
	   {
	     vl = &(x1[24 * i + 6 * k]);
	     vr = &(x2[24 * i + 6 * k]);
	     v =  &(x3[24 * i + 6 * k]);

	     for(l = 0; l < 6; l++)
	       v[l] = 0;

	     for(l = 0; l < 6; l++)
	       {
		 al = 0.0;
		 ar = 0.0;
		 for(j = 0; j < 6; j++)
		   {
		     al += vl[j] * left[k * 36 + l * 6 + j];
		     ar += vr[j] * right[k * 36 + l * 6 + j];
		   }

		 x1px2 = al * ar;
		 for(j = 0; j < 6; j++)
		   v[j] += x1px2 * extEV[6 * l + j];
	       }
	   }
	 
	 v = &(x3[24 * i]);
	 scale = 1;
	 for(l = 0; scale && (l < 24); l++)
	   scale = ((ABS(v[l]) <  minlikelihood));

	 if (scale)
	   {
	     for(l = 0; l < 24; l++)
	       v[l] *= twotothe256;

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;	   
	   }
       }
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;

}


static void newviewGTRGAMMASECONDARY_7(int tipCase,
				       double *x1, double *x2, double *x3, double *extEV, double *tipVector,
				       int *ex3, unsigned char *tipX1, unsigned char *tipX2,
				       int n, double *left, double *right, int *wgt, int *scalerIncrement, const boolean useFastScaling)
{
  double  *v;
  double x1px2;
  int  i, j, l, k, scale, addScale = 0;
  double *vl, *vr, al, ar;

  switch(tipCase)
    {
    case TIP_TIP:
      {
	for(i = 0; i < n; i++)
	  {
	    for(k = 0; k < 4; k++)
	      {
		vl = &(tipVector[7 * tipX1[i]]);
		vr = &(tipVector[7 * tipX2[i]]);
		v =  &(x3[28 * i + 7 * k]);

		for(l = 0; l < 7; l++)
		  v[l] = 0;

		for(l = 0; l < 7; l++)
		  {
		    al = 0.0;
		    ar = 0.0;
		    for(j = 0; j < 7; j++)
		      {
			al += vl[j] * left[k * 49 + l * 7 + j];
			ar += vr[j] * right[k * 49 + l * 7 + j];
		      }

		    x1px2 = al * ar;
		    for(j = 0; j < 7; j++)
		      v[j] += x1px2 * extEV[7 * l + j];
		  }
	      }	   
	  }
      }
      break;
    case TIP_INNER:
      {
	for (i = 0; i < n; i++)
	  {
	    for(k = 0; k < 4; k++)
	      {
		vl = &(tipVector[7 * tipX1[i]]);
		vr = &(x2[28 * i + 7 * k]);
		v =  &(x3[28 * i + 7 * k]);

		for(l = 0; l < 7; l++)
		  v[l] = 0;

		for(l = 0; l < 7; l++)
		  {
		    al = 0.0;
		    ar = 0.0;
		    for(j = 0; j < 7; j++)
		      {
			al += vl[j] * left[k * 49 + l * 7 + j];
			ar += vr[j] * right[k * 49 + l * 7 + j];
		      }

		    x1px2 = al * ar;
		    for(j = 0; j < 7; j++)
		      v[j] += x1px2 * extEV[7 * l + j];
		  }
	      }
	   
	    v = &x3[28 * i];
	    scale = 1;
	    for(l = 0; scale && (l < 28); l++)
	      scale = (ABS(v[l]) <  minlikelihood);

	    if (scale)
	      {
		for(l = 0; l < 28; l++)
		  v[l] *= twotothe256;

		if(useFastScaling)
		  addScale += wgt[i];
		else
		  ex3[i]  += 1;	      
	      }
	  }
      }
      break;
    case INNER_INNER:
      for (i = 0; i < n; i++)
       {
	 for(k = 0; k < 4; k++)
	   {
	     vl = &(x1[28 * i + 7 * k]);
	     vr = &(x2[28 * i + 7 * k]);
	     v =  &(x3[28 * i + 7 * k]);

	     for(l = 0; l < 7; l++)
	       v[l] = 0;

	     for(l = 0; l < 7; l++)
	       {
		 al = 0.0;
		 ar = 0.0;
		 for(j = 0; j < 7; j++)
		   {
		     al += vl[j] * left[k * 49 + l * 7 + j];
		     ar += vr[j] * right[k * 49 + l * 7 + j];
		   }

		 x1px2 = al * ar;
		 for(j = 0; j < 7; j++)
		   v[j] += x1px2 * extEV[7 * l + j];
	       }
	   }
	 
	 v = &(x3[28 * i]);
	 scale = 1;
	 for(l = 0; scale && (l < 28); l++)
	   scale = ((ABS(v[l]) <  minlikelihood));

	 if (scale)
	   {
	     for(l = 0; l < 28; l++)
	       v[l] *= twotothe256;

	     if(useFastScaling)
	       addScale += wgt[i];
	     else
	       ex3[i]  += 1;	   
	   }
       }
      break;
    default:
      assert(0);
    }

  if(useFastScaling)
    *scalerIncrement = addScale;
}







void computeTraversalInfo(nodeptr p, traversalInfo *ti, int *counter, int maxTips, int numBranches)
{
  if(isTip(p->number, maxTips))
    return;

  {
    int i;
    nodeptr q = p->next->back;
    nodeptr r = p->next->next->back;

    if(isTip(r->number, maxTips) && isTip(q->number, maxTips))
      {
	while (! p->x)
	 {
	   if (! p->x)
	     getxnode(p);
	 }

	ti[*counter].tipCase = TIP_TIP;
	ti[*counter].pNumber = p->number;
	ti[*counter].qNumber = q->number;
	ti[*counter].rNumber = r->number;
	for(i = 0; i < numBranches; i++)
	  {
	    double z;
	    z = q->z[i];
	    z = (z > zmin) ? log(z) : log(zmin);
	    ti[*counter].qz[i] = z;

	    z = r->z[i];
	    z = (z > zmin) ? log(z) : log(zmin);
	    ti[*counter].rz[i] = z;
	  }
	*counter = *counter + 1;
      }
    else
      {
	if(isTip(r->number, maxTips) || isTip(q->number, maxTips))
	  {
	    nodeptr tmp;

	    if(isTip(r->number, maxTips))
	      {
		tmp = r;
		r = q;
		q = tmp;
	      }

	    while ((! p->x) || (! r->x))
	      {
		if (! r->x)
		  computeTraversalInfo(r, ti, counter, maxTips, numBranches);
		if (! p->x)
		  getxnode(p);
	      }

	    ti[*counter].tipCase = TIP_INNER;
	    ti[*counter].pNumber = p->number;
	    ti[*counter].qNumber = q->number;
	    ti[*counter].rNumber = r->number;
	    for(i = 0; i < numBranches; i++)
	      {
		double z;
		z = q->z[i];
		z = (z > zmin) ? log(z) : log(zmin);
		ti[*counter].qz[i] = z;

		z = r->z[i];
		z = (z > zmin) ? log(z) : log(zmin);
		ti[*counter].rz[i] = z;
	      }

	    *counter = *counter + 1;
	  }
	else
	  {

	    while ((! p->x) || (! q->x) || (! r->x))
	      {
		if (! q->x)
		  computeTraversalInfo(q, ti, counter, maxTips, numBranches);
		if (! r->x)
		  computeTraversalInfo(r, ti, counter, maxTips, numBranches);
		if (! p->x)
		  getxnode(p);
	      }

	    ti[*counter].tipCase = INNER_INNER;
	    ti[*counter].pNumber = p->number;
	    ti[*counter].qNumber = q->number;
	    ti[*counter].rNumber = r->number;
	    for(i = 0; i < numBranches; i++)
	      {
		double z;
		z = q->z[i];
		z = (z > zmin) ? log(z) : log(zmin);
		ti[*counter].qz[i] = z;

		z = r->z[i];
		z = (z > zmin) ? log(z) : log(zmin);
		ti[*counter].rz[i] = z;
	      }

	    *counter = *counter + 1;
	  }
      }
  }

  

}






void newviewIterative (tree *tr)
{
  traversalInfo *ti   = tr->td[0].ti;
  int i, model;

  for(i = 1; i < tr->td[0].count; i++)
    {
      traversalInfo *tInfo = &ti[i];

      for(model = 0; model < tr->NumberOfModels; model++)
	{
	  if(tr->executeModel[model])
	    {	      
	      double
		*x1_start = (double*)NULL,
		*x2_start = (double*)NULL,
		*x3_start = (double*)NULL,
		*left     = (double*)NULL,
		*right    = (double*)NULL;
	      float
		*x1_start_FLOAT = (float*)NULL,
		*x2_start_FLOAT = (float*)NULL,
		*x3_start_FLOAT = (float*)NULL,
		*left_FLOAT     = (float*)NULL,
		*right_FLOAT    = (float*)NULL;
	      int		
		scalerIncrement = 0,
		*wgt = (int*)NULL,	       
		*ex3 = (int*)NULL;
	      unsigned char
		*tipX1 = (unsigned char *)NULL,
		*tipX2 = (unsigned char *)NULL;
	      double qz, rz;
	      int width =  tr->partitionData[model].width;

	      if(tr->useFastScaling)		
		wgt   =  tr->partitionData[model].wgt;		 		  				

	      switch(tInfo->tipCase)
		{
		case TIP_TIP:
		  tipX1    = tr->partitionData[model].yVector[tInfo->qNumber];
		  tipX2    = tr->partitionData[model].yVector[tInfo->rNumber];

		  if(!tr->useFloat)
		    x3_start       = tr->partitionData[model].xVector[tInfo->pNumber - tr->mxtips - 1];
		  else
		    x3_start_FLOAT = tr->partitionData[model].xVector_FLOAT[tInfo->pNumber - tr->mxtips - 1];

		  if(!tr->useFastScaling)
		    {
		      int k;
		      ex3      = tr->partitionData[model].expVector[tInfo->pNumber - tr->mxtips - 1];

		      for(k = 0; k < width; k++)
			ex3[k] = 0;
		    }

		  break;
		case TIP_INNER:
		  tipX1    =  tr->partitionData[model].yVector[tInfo->qNumber];

		  if(!tr->useFloat)
		    {
		      x2_start       = tr->partitionData[model].xVector[tInfo->rNumber - tr->mxtips - 1];
		      x3_start       = tr->partitionData[model].xVector[tInfo->pNumber - tr->mxtips - 1];
		    }
		  else
		    {
		      x2_start_FLOAT = tr->partitionData[model].xVector_FLOAT[tInfo->rNumber - tr->mxtips - 1];		 		    		 
		      x3_start_FLOAT =  tr->partitionData[model].xVector_FLOAT[tInfo->pNumber - tr->mxtips - 1];
		    }

		  if(!tr->useFastScaling)
		    {
		      int 
			k,
			*ex2;
		      
		      ex2      = tr->partitionData[model].expVector[tInfo->rNumber - tr->mxtips - 1];
		      ex3      = tr->partitionData[model].expVector[tInfo->pNumber - tr->mxtips - 1];
		      
		      for(k = 0; k < width; k++)
			ex3[k] = ex2[k];
		    }

		  break;
		case INNER_INNER:

		  if(!tr->useFloat)
		    {
		      x1_start       = tr->partitionData[model].xVector[tInfo->qNumber - tr->mxtips - 1];
		      x2_start       = tr->partitionData[model].xVector[tInfo->rNumber - tr->mxtips - 1];
		      x3_start       = tr->partitionData[model].xVector[tInfo->pNumber - tr->mxtips - 1];
		    }		      
		  else
		    {
		      x1_start_FLOAT = tr->partitionData[model].xVector_FLOAT[tInfo->qNumber - tr->mxtips - 1];		  
		      x2_start_FLOAT = tr->partitionData[model].xVector_FLOAT[tInfo->rNumber - tr->mxtips - 1];		 
		      x3_start_FLOAT = tr->partitionData[model].xVector_FLOAT[tInfo->pNumber - tr->mxtips - 1];
		    }

		  if(!tr->useFastScaling)
		    {
		      int 
			k,
			*ex1,
			*ex2;

		      ex1      = tr->partitionData[model].expVector[tInfo->qNumber - tr->mxtips - 1];
		      ex2      = tr->partitionData[model].expVector[tInfo->rNumber - tr->mxtips - 1];
		      ex3      = tr->partitionData[model].expVector[tInfo->pNumber - tr->mxtips - 1];
		      
		      for(k = 0; k < width; k++)
			ex3[k] = ex1[k] + ex2[k];
		    }		  
		  break;
		default:
		  assert(0);
		}

	      if(tr->useFloat)
		{
		  left_FLOAT = tr->partitionData[model].left_FLOAT;
		  right_FLOAT = tr->partitionData[model].right_FLOAT;
		}
	      else
		{
		  left  = tr->partitionData[model].left;
		  right = tr->partitionData[model].right;
		}

	      if(tr->multiBranch)
		{
		  qz = tInfo->qz[model];
		  rz = tInfo->rz[model];
		}
	      else
		{
		  qz = tInfo->qz[0];
		  rz = tInfo->rz[0];
		}


	      switch(tr->partitionData[model].dataType)
		{
		case BINARY_DATA:
		  switch(tr->rateHetModel)
		    {
		    case CAT:
		      {			
			makeP(qz, rz, tr->cdta->patrat,   tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN, tr->NumberOfCategories,
			      left, right, BINARY_DATA);

			newviewGTRCAT_BINARY(tInfo->tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
					     x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
					     ex3, tipX1, tipX2,
					     width, left, right, wgt, &scalerIncrement, tr->useFastScaling
					     );		
		      }
		      break;
		    case GAMMA:
		    case GAMMA_I:
		      {
			makeP(qz, rz, tr->partitionData[model].gammaRates,
			      tr->partitionData[model].EI, tr->partitionData[model].EIGN,
			      4, left, right, BINARY_DATA);

			newviewGTRGAMMA_BINARY(tInfo->tipCase,
					       x1_start, x2_start, x3_start, tr->partitionData[model].EV, tr->partitionData[model].tipVector,
					       ex3, tipX1, tipX2,
					       width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
		      }
		      break;
		    default:
		      assert(0);
		    }
		  break;
		case DNA_DATA:
		  switch(tr->rateHetModel)
		    {
		    case CAT:
		      if(tr->useFloat)
			{			  
			   makeP_FLOAT(qz, rz, tr->cdta->patrat,   tr->partitionData[model].EI,
				       tr->partitionData[model].EIGN, tr->NumberOfCategories,
				       left_FLOAT, right_FLOAT, DNA_DATA);


			   newviewGTRCAT_FLOAT(tInfo->tipCase,  tr->partitionData[model].EV_FLOAT, tr->partitionData[model].rateCategory,
					       x1_start_FLOAT, x2_start_FLOAT, x3_start_FLOAT, tr->partitionData[model].tipVector_FLOAT,
					       ex3, tipX1, tipX2,
					       width, left_FLOAT, right_FLOAT, wgt, &scalerIncrement, tr->useFastScaling
					       );			  
			}
		      else
			{			  
			  makeP(qz, rz, tr->cdta->patrat,   tr->partitionData[model].EI,
				tr->partitionData[model].EIGN, tr->NumberOfCategories,
				left, right, DNA_DATA);

			  newviewGTRCAT(tInfo->tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
					x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
					ex3, tipX1, tipX2,
					width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
			}
		      break;
		    case GAMMA:
		    case GAMMA_I:
		      if(tr->useFloat)
			{
			  makeP_FLOAT(qz, rz, tr->partitionData[model].gammaRates,
				      tr->partitionData[model].EI, tr->partitionData[model].EIGN,
				      4, left_FLOAT, right_FLOAT, DNA_DATA);

			  newviewGTRGAMMA_FLOAT(tInfo->tipCase,
						x1_start_FLOAT, x2_start_FLOAT, x3_start_FLOAT, tr->partitionData[model].EV_FLOAT, tr->partitionData[model].tipVector_FLOAT,
						ex3, tipX1, tipX2,
						width, left_FLOAT, right_FLOAT, wgt, &scalerIncrement, tr->useFastScaling);			

			}
		      else
			{			 
			  makeP(qz, rz, tr->partitionData[model].gammaRates,
				tr->partitionData[model].EI, tr->partitionData[model].EIGN,
				4, left, right, DNA_DATA);

			  newviewGTRGAMMA(tInfo->tipCase,
					  x1_start, x2_start, x3_start, tr->partitionData[model].EV, tr->partitionData[model].tipVector,
					  ex3, tipX1, tipX2,
					  width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
			}
		      break;
		    default:
		      assert(0);
		    }
		  break;
		case AA_DATA:
		  switch(tr->rateHetModel)
		    {
		    case CAT:
		      {
			makeP(qz, rz, tr->cdta->patrat,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      tr->NumberOfCategories, left, right, AA_DATA);

			newviewGTRCATPROT(tInfo->tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
					  x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
					  ex3, tipX1, tipX2, width, left, right, wgt, &scalerIncrement, tr->useFastScaling);			
		      }
		      break;
		    case GAMMA:
		    case GAMMA_I:
		      if(tr->useFloat)
			{
			  makeP_FLOAT(qz, rz, tr->partitionData[model].gammaRates,
				      tr->partitionData[model].EI,
				      tr->partitionData[model].EIGN,
				      4, left_FLOAT, right_FLOAT, AA_DATA);

			  newviewGTRGAMMAPROT_FLOAT(tInfo->tipCase,
						    x1_start_FLOAT, x2_start_FLOAT, x3_start_FLOAT,
						    tr->partitionData[model].EV_FLOAT,
						    tr->partitionData[model].tipVector_FLOAT,
						    ex3, tipX1, tipX2,
						    width, left_FLOAT, right_FLOAT, wgt, &scalerIncrement, tr->useFastScaling);
			}
		      else
			{
			  makeP(qz, rz, tr->partitionData[model].gammaRates,
				tr->partitionData[model].EI,
				tr->partitionData[model].EIGN,
				4, left, right, AA_DATA);

			  newviewGTRGAMMAPROT(tInfo->tipCase,
					      x1_start, x2_start, x3_start,
					      tr->partitionData[model].EV,
					      tr->partitionData[model].tipVector,
					      ex3, tipX1, tipX2,
					      width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
			}
		      break;
		    default:
		      assert(0);
		    }
		  break;
		case SECONDARY_DATA_6:
		  switch(tr->rateHetModel)
		    {
		    case CAT:
		      {
			makeP(qz, rz, tr->cdta->patrat,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      tr->NumberOfCategories, left, right, SECONDARY_DATA_6);

			newviewGTRCATSECONDARY_6(tInfo->tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
						 x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
						 ex3, tipX1, tipX2, width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
		      }
		      break;
		    case GAMMA:
		    case GAMMA_I:
		      {
			makeP(qz, rz, tr->partitionData[model].gammaRates,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      4, left, right, SECONDARY_DATA_6);

			newviewGTRGAMMASECONDARY_6(tInfo->tipCase,
						   x1_start, x2_start, x3_start,
						   tr->partitionData[model].EV,
						   tr->partitionData[model].tipVector,
						   ex3, tipX1, tipX2,
						   width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
		      }
		      break;
		    default:
		      assert(0);
		    }
		  break;
		case SECONDARY_DATA_7:
		  switch(tr->rateHetModel)
		    {
		    case CAT:
		      {
			makeP(qz, rz, tr->cdta->patrat,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      tr->NumberOfCategories, left, right, SECONDARY_DATA_7);

			newviewGTRCATSECONDARY_7(tInfo->tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
						 x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
						 ex3, tipX1, tipX2, width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
		      }
		      break;
		    case GAMMA:
		    case GAMMA_I:
		      {
			makeP(qz, rz, tr->partitionData[model].gammaRates,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      4, left, right, SECONDARY_DATA_7);

			newviewGTRGAMMASECONDARY_7(tInfo->tipCase,
						   x1_start, x2_start, x3_start,
						   tr->partitionData[model].EV,
						   tr->partitionData[model].tipVector,
						   ex3, tipX1, tipX2,
						   width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
		      }
		      break;
		    default:
		      assert(0);
		    }
		  break;
		case SECONDARY_DATA:
		  switch(tr->rateHetModel)
		    {
		    case CAT:
		      {
			makeP(qz, rz, tr->cdta->patrat,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      tr->NumberOfCategories, left, right, SECONDARY_DATA);

			newviewGTRCATSECONDARY(tInfo->tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
					       x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
					       ex3, tipX1, tipX2, width, left, right, wgt, &scalerIncrement, tr->useFastScaling);
		      }
		      break;
		    case GAMMA:
		    case GAMMA_I:
		      {
			makeP(qz, rz, tr->partitionData[model].gammaRates,
			      tr->partitionData[model].EI,
			      tr->partitionData[model].EIGN,
			      4, left, right, SECONDARY_DATA);

			newviewGTRGAMMASECONDARY(tInfo->tipCase,
						 x1_start, x2_start, x3_start,
						 tr->partitionData[model].EV,
						 tr->partitionData[model].tipVector,
						 ex3, tipX1, tipX2,
						 width, left, right, wgt, &scalerIncrement, tr->useFastScaling);		
		      }
		      break;
		    default:
		      assert(0);
		    }
		  break;
		default:
		  assert(0);
		}
	      if(tr->useFastScaling)
		{
		  tr->partitionData[model].globalScaler[tInfo->pNumber] = 
		    tr->partitionData[model].globalScaler[tInfo->qNumber] + 
		    tr->partitionData[model].globalScaler[tInfo->rNumber] +
		    (unsigned int)scalerIncrement;
		  assert(tr->partitionData[model].globalScaler[tInfo->pNumber] < INT_MAX);
		}
	    }
	}



    }
}

void newviewGeneric (tree *tr, nodeptr p)
{  
  if(isTip(p->number, tr->mxtips))
    return;


  {
    tr->td[0].count = 1;
    computeTraversalInfo(p, &(tr->td[0].ti[0]), &(tr->td[0].count), tr->mxtips, tr->numBranches);

    if(tr->td[0].count > 1)
      {
#ifdef _USE_PTHREADS
	masterBarrier(THREAD_NEWVIEW, tr);
#else
	newviewIterative(tr);
#endif
      }
  }
}


void newviewGenericMasked(tree *tr, nodeptr p)
{
  if(isTip(p->number, tr->mxtips))
    return;

  {
    int i;

    for(i = 0; i < tr->NumberOfModels; i++)
      {
	if(tr->partitionConverged[i])
	  tr->executeModel[i] = FALSE;
	else
	  tr->executeModel[i] = TRUE;
      }

    tr->td[0].count = 1;
    computeTraversalInfo(p, &(tr->td[0].ti[0]), &(tr->td[0].count), tr->mxtips, tr->numBranches);

    if(tr->td[0].count > 1)
      {
#ifdef _USE_PTHREADS
	masterBarrier(THREAD_NEWVIEW_MASKED, tr);
#else
	newviewIterative(tr);
#endif
      }

    for(i = 0; i < tr->NumberOfModels; i++)
      tr->executeModel[i] = TRUE;
  }
}


#ifdef  _USE_PTHREADS_MULTIGRAIN

void newviewClassify (tree *tr, branchInfo *b, double *z)
{
  int
    *wgt = (int*)NULL,
    scalerIncrement = 0,
    k,
    model,
    leftNumber = b->leftNodeNumber,
    rightNumber = b->rightNodeNumber,
    tipCase,
    increment,
    columnCounter = 0,
    offsetCounter = 0;

  assert(!tr->useFastScaling && !tr->useFloat);

  for(model = 0; model < tr->NumberOfModels; model++)
    {
      double
	*x1_start = (double*)NULL,
	*x2_start = (double*)NULL,
	*x3_start = &(tr->temporaryVector[offsetCounter]),
	*left     = tr->partitionData[model].left,
	*right    = tr->partitionData[model].right,
	pz;

      int
	*ex1 = (int*)NULL,
	*ex2 = (int*)NULL,
	*ex3 = &(tr->temporaryScaling[columnCounter]),
	width = tr->partitionData[model].upper - tr->partitionData[model].lower;

      unsigned char
	*tipX1 = (unsigned char *)NULL,
	*tipX2 = (unsigned char *)NULL;     

      if (isTip(leftNumber, tr->mxtips) && isTip(rightNumber, tr->mxtips))
	{
	  tipCase = TIP_TIP;

	  tipX1 = &tr->contiguousTips[leftNumber][columnCounter];
	  tipX2 = &tr->contiguousTips[rightNumber][columnCounter];

	  for(k = 0; k < width; k++)
	    ex3[k] = 0;

	}
      else
	{
	  if (isTip(leftNumber, tr->mxtips))
	    {
	      tipCase = TIP_INNER;

	      tipX1 = &tr->contiguousTips[leftNumber][columnCounter];

	      x2_start = &b->right[offsetCounter];
	      ex2      = &b->rightScaling[columnCounter];

	      for(k = 0; k < width; k++)
		ex3[k] = ex2[k];

	    }
	  else
	    {
	      if(isTip(rightNumber, tr->mxtips))
		{
		  tipCase = TIP_INNER;

		  tipX1 = &tr->contiguousTips[rightNumber][columnCounter];

		  x2_start = &b->left[offsetCounter];
		  ex2      = &b->leftScaling[columnCounter];

		  for(k = 0; k < width; k++)
		    ex3[k] = ex2[k];

		}
	      else
		{
		  tipCase = INNER_INNER;

		  x1_start = &b->left[offsetCounter];
		  ex1      = &b->leftScaling[columnCounter];

		  x2_start = &b->right[offsetCounter];
		  ex2      = &b->rightScaling[columnCounter];

		  for(k = 0; k < width; k++)
		    ex3[k] = ex1[k] + ex2[k];

		}
	    }
	}     

      if(tr->multiBranch)
	{
	  pz = z[model];
	  pz = (pz > zmin) ? log(pz) : log(zmin);
	}
      else
	{
	  pz = z[0];
	  pz = (pz > zmin) ? log(pz) : log(zmin);
	}


      switch(tr->partitionData[model].dataType)
	{
	case BINARY_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:
	      {	
		makeP(pz, pz, tr->cdta->patrat,   tr->partitionData[model].EI,
		      tr->partitionData[model].EIGN, tr->NumberOfCategories,
		      left, right, BINARY_DATA);

		newviewGTRCAT_BINARY(tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
				     x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
				     ex3, tipX1, tipX2,
				     width, left, right, wgt, &scalerIncrement, tr->useFastScaling
				     );

		increment = 2;	
	      }
	      break;
	    case GAMMA:
	    case GAMMA_I:
	      {	
		makeP(pz, pz, tr->partitionData[model].gammaRates,
		      tr->partitionData[model].EI, tr->partitionData[model].EIGN,
		      4, left, right, BINARY_DATA);

		newviewGTRGAMMA_BINARY(tipCase,
				       x1_start, x2_start, x3_start, tr->partitionData[model].EV, tr->partitionData[model].tipVector,
				       ex3, tipX1, tipX2,
				       width, left, right, wgt, &scalerIncrement, tr->useFastScaling);

		increment = 8;		
	      }
	      break;
	    default:
	      assert(0);
	    }
	  break;
	case DNA_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:
	      {	
		makeP(pz, pz, tr->cdta->patrat,   tr->partitionData[model].EI,
		      tr->partitionData[model].EIGN, tr->NumberOfCategories,
		      left, right, DNA_DATA);

		newviewGTRCAT(tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
			      x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
			      ex3, tipX1, tipX2,
			      width, left, right, wgt, &scalerIncrement, tr->useFastScaling
			      );

		increment = 4;	
	      }
	      break;
	    case GAMMA:
	    case GAMMA_I:
	      {		
		makeP(pz, pz, tr->partitionData[model].gammaRates,
		      tr->partitionData[model].EI, tr->partitionData[model].EIGN,
		      4, left, right, DNA_DATA);

		newviewGTRGAMMA(tipCase,
				x1_start, x2_start, x3_start, tr->partitionData[model].EV, tr->partitionData[model].tipVector,
				ex3, tipX1, tipX2,
				width, left, right, wgt, &scalerIncrement, tr->useFastScaling);

		increment = 16;	
	      }
	      break;
	    default:
	      assert(0);
	    }
	  break;
	case AA_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:
	      {
		makeP(pz, pz, tr->cdta->patrat,
		      tr->partitionData[model].EI,
		      tr->partitionData[model].EIGN,
		      tr->NumberOfCategories, left, right, AA_DATA);

		newviewGTRCATPROT(tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
				  x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
				  ex3, tipX1, tipX2, width, left, right, wgt, &scalerIncrement, tr->useFastScaling);

		increment = 20;	       
	      }
	      break;
	    case GAMMA:
	    case GAMMA_I:
	      {	
		makeP(pz, pz, tr->partitionData[model].gammaRates,
		      tr->partitionData[model].EI,
		      tr->partitionData[model].EIGN,
		      4, left, right, AA_DATA);

		newviewGTRGAMMAPROT(tipCase,
				    x1_start, x2_start, x3_start,
				    tr->partitionData[model].EV,
				    tr->partitionData[model].tipVector,
				    ex3, tipX1, tipX2,
				    width, left, right, wgt, &scalerIncrement, tr->useFastScaling);

		increment = 80;		
	      }
	      break;
	    default:
	      assert(0);
	    }
	  break;
	case SECONDARY_DATA:
	  switch(tr->rateHetModel)
	    {
	    case CAT:
	      {	
		makeP(pz, pz, tr->cdta->patrat,
		      tr->partitionData[model].EI,
		      tr->partitionData[model].EIGN,
		      tr->NumberOfCategories, left, right, SECONDARY_DATA);

		newviewGTRCATSECONDARY(tipCase,  tr->partitionData[model].EV, tr->partitionData[model].rateCategory,
				       x1_start, x2_start, x3_start, tr->partitionData[model].tipVector,
				       ex3, tipX1, tipX2, width, left, right, wgt, &scalerIncrement, tr->useFastScaling);

		increment = 16;		
	      }
	      break;
	    case GAMMA:
	    case GAMMA_I:
	      {		
		makeP(pz, pz, tr->partitionData[model].gammaRates,
		      tr->partitionData[model].EI,
		      tr->partitionData[model].EIGN,
		      4, left, right, SECONDARY_DATA);

		newviewGTRGAMMASECONDARY(tipCase,
					 x1_start, x2_start, x3_start,
					 tr->partitionData[model].EV,
					 tr->partitionData[model].tipVector,
					 ex3, tipX1, tipX2,
					 width, left, right, wgt, &scalerIncrement, tr->useFastScaling);

		increment = 64;	
	      }
	      break;
	    default:
	      assert(0);
	    }
	  break;
	default:
	  assert(0);
	}
      columnCounter += width;
      offsetCounter += width * increment;
    }
}


void newviewClassifySpecial(tree *tr, double *x1_start, double *x2_start, double *x3_start, int *ex1, int *ex2, int *ex3,
			    unsigned char *tipX1,  unsigned char *tipX2, int tipCase, double pz, double qz)
{
  int 
    k,
    *wgt = (int*)NULL,
    scalerIncrement = 0;

  double
    _pz = (pz > zmin) ? log(pz) : log(zmin),
    _qz = (qz > zmin) ? log(qz) : log(zmin);

  double *left  = (double *)malloc_aligned(64 * sizeof(double));
  double *right = (double *)malloc_aligned(64 * sizeof(double));

  assert(!tr->useFastScaling);

  int width = tr->partitionData[0].upper - tr->partitionData[0].lower;

  switch(tipCase)
    {
    case TIP_TIP:
      for(k = 0; k < width; k++)
	ex3[k] = 0;
      break;
    case TIP_INNER:
      for(k = 0; k < width; k++)
	ex3[k] = ex2[k];
      break;
    case INNER_INNER:
       for(k = 0; k < width; k++)
	ex3[k] = ex1[k] + ex2[k];
      break;
    default:
      assert(0);      
    }

  makeP(_pz, _qz, tr->partitionData[0].gammaRates,
	tr->partitionData[0].EI, tr->partitionData[0].EIGN,
	4, left, right, DNA_DATA);

  newviewGTRGAMMA(tipCase,
		  x1_start, x2_start, x3_start, tr->partitionData[0].EV, tr->partitionData[0].tipVector,
		  ex3, tipX1, tipX2,
		  width, left, right, wgt, &scalerIncrement, tr->useFastScaling);


  free(left);
  free(right);
}




#endif


